package antlr;

/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/license.html
 *
 * $Id: $
 */

//
// ANTLR C# Code Generator by Micheal Jordan
//                            Kunle Odutola       : kunle UNDERSCORE odutola AT hotmail DOT com
//                            Anthony Oguntimehin
//
// With many thanks to Eric V. Smith from the ANTLR list.
//

// HISTORY:
//
// 17-May-2002 kunle    Fixed bug in OctalToUnicode() - was processing non-Octal escape sequences
//                      Also added namespace support based on Cpp version.
// 07-Jun-2002 kunle    Added Scott Ellis's _saveIndex creation optimizations
// 09-Sep-2002 richardN Richard Ney's bug-fix for literals table construction.
//                      [ Hashtable ctor needed instance of hash code provider not it's class name. ]
// 17-Sep-2002 kunle &  Added all Token ID definitions as data member of every Lexer/Parser/TreeParser
//             AOg      [ A by-product of problem-solving phase of the hetero-AST changes below
//                        but, it breaks nothing and restores "normal" ANTLR codegen behaviour. ]
// 19-Oct-2002 kunle &  Completed the work required to support heterogenous ASTs (many changes)
//             AOg   &
//             michealj
// 14-Nov-2002 michealj Added "initializeASTFactory()" to support flexible ASTFactory initialization.
//						[ Thanks to Ric Klaren - for suggesting it and implementing it for Cpp. ]
// 18-Nov-2002 kunle    Added fix to make xx_tokenSet_xx names CLS compliant.
// 01-Dec-2002 richardN Patch to reduce "unreachable code" warnings
// 01-Dec-2002 richardN Fix to generate correct TreeParser token-type classnames.
// 12-Jan-2003 kunle  & Generated Lexers, Parsers and TreeParsers now support ANTLR's tracing option.
//             michealj
// 12-Jan-2003 kunle    Fixed issue where initializeASTFactory() was generated when "buildAST=false"
// 14-Jan-2003 AOg      initializeASTFactory(AST factory) method was modifying the Parser's "astFactory"
//                      member rather than it's own "factory" parameter. Fixed.
// 18-Jan-2003 kunle  & Fixed reported issues with ASTFactory create() calls for hetero ASTs
//             michealj - code generated for LEXER token with hetero-AST option specified does not compile
//                      - code generated for imaginary tokens with hetero-AST option specified uses
//                        default AST type
//                      - code generated for per-TokenRef hetero-AST option specified does not compile
// 18-Jan-2003 kunle    initializeASTFactory(AST) method is now a static public member
// 18-May-2003 kunle    Changes to address outstanding reported issues::
//                      - Fixed reported issues with support for case-sensitive literals
//                      - antlr.SemanticException now imported for all Lexers.
//                        [ This exception is thrown on predicate failure. ]
// 12-Jan-2004 kunle    Added fix for reported issue with un-compileable generated lexers
//
//

import java.util.Enumeration;
import java.util.Hashtable;
import antlr.collections.impl.BitSet;
import antlr.collections.impl.Vector;
import java.io.PrintWriter; //SAS: changed for proper text file io
import java.io.IOException;
import java.io.FileWriter;
import java.util.StringTokenizer;
import java.text.*;

/** Generates MyParser.vb, MyLexer.vb and MyParserTokenTypes.vb */
public class VBCodeGenerator
    extends CodeGenerator {
    public static String newline = System.getProperty("line.separator");
    protected String tabstr = "";
    protected void tabadd() {
        tabs++;
        tabstr += "\t";
    }

    protected void tabsub() {
        tabs--;
        int pos = tabstr.lastIndexOf("\t");
        if (pos > 0) {
            tabstr = tabstr.substring(0, pos);
        }
    }

    protected void tabclear() {
        tabs = 0;
        tabstr = "";
    }


    // non-zero if inside syntactic predicate generation
    protected int syntacticPredLevel = 0;

    // Are we generating ASTs (for parsers and tree parsers) right now?
    protected boolean genAST = false;

    // Are we saving the text consumed (for lexers) right now?
    protected boolean saveText = false;

    // Grammar parameters set up to handle different grammar classes.
    // These are used to get instanceof tests out of code generation
    boolean usingCustomAST = false;
    String labeledElementType;
    String labeledElementASTType;
    String labeledElementInit;
    String commonExtraArgs;
    String commonExtraParams;
    String commonLocalVars;
    String lt1Value;
    String exceptionThrown;
    String throwNoViable;

    // Tracks the rule being generated.  Used for mapTreeId
    RuleBlock currentRule;
    // Tracks the rule or labeled subrule being generated.  Used for AST generation.
    String currentASTResult;

    /** Mapping between the ids used in the current alt, and the
     * names of variables used to represent their AST values.
     */
    Hashtable treeVariableMap = new Hashtable();

    /** Used to keep track of which AST variables have been defined in a rule
     * (except for the #rule_name and #rule_name_in var's
     */
    Hashtable declaredASTVariables = new Hashtable();

    /* Count of unnamed generated variables */
    int astVarNumber = 1;

    /** Special value used to mark duplicate in treeVariableMap */
    protected static final String NONUNIQUE = new String();

    public static final int caseSizeThreshold = 127; // ascii is max

    private Vector semPreds;
    // Used to keep track of which (heterogeneous AST types are used)
    // which need to be set in the ASTFactory of the generated parser
    private java.util.Vector astTypes;

    private static VBNameSpace nameSpace = null;

    // _saveIndex creation optimization -- don't create it unless we need to use it
    boolean bSaveIndexCreated = false;

    /** Create a vb code-generator using the given Grammar.
     * The caller must still call setTool, setBehavior, and setAnalyzer
     * before generating code.
     */
    public VBCodeGenerator() {
        super();
        charFormatter = new VBCharFormatter();
    }

    /** Adds a semantic predicate string to the sem pred vector
        These strings will be used to build an array of sem pred names
        when building a debugging parser.  This method should only be
        called when the debug option is specified
     */
    protected int addSemPred(String predicate) {
        semPreds.appendElement(predicate);
        return semPreds.size() - 1;
    }

    public void exitIfError() {
        if (antlrTool.hasError()) {
            antlrTool.fatalError("Exiting due to errors.");
        }
    }

    /**Generate the parser, lexer, treeparser, and token types in VB */
    public void gen() {
        // Do the code generation
        try {
            // Loop over all grammars
            Enumeration grammarIter = behavior.grammars.elements();
            while (grammarIter.hasMoreElements()) {
                Grammar g = (Grammar) grammarIter.nextElement();
                // Connect all the components to each other
                g.setGrammarAnalyzer(analyzer);
                g.setCodeGenerator(this);
                analyzer.setGrammar(g);
                // To get right overloading behavior across heterogeneous grammars
                setupGrammarParameters(g);
                g.generate();
                exitIfError();
            }

            // Loop over all token managers (some of which are lexers)
            Enumeration tmIter = behavior.tokenManagers.elements();
            while (tmIter.hasMoreElements()) {
                TokenManager tm = (TokenManager) tmIter.nextElement();
                if (!tm.isReadOnly()) {
                    // Write the token manager tokens as vb
                    // this must appear before genTokenInterchange so that
                    // labels are set on string literals
                    genTokenTypes(tm);
                    // Write the token manager tokens as plain text
                    genTokenInterchange(tm);
                }
                exitIfError();
            }
        }
        catch (IOException e) {
            antlrTool.reportException(e, null);
        }
    }

    /** Generate code for the given grammar element.
     * @param blk The {...} action to generate
     */
    public void gen(ActionElement action) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genAction(" + action + ")");
        if (action.isSemPred) {

            genSemPred(action.actionText, action.line);
        }
        else {
            if (grammar.hasSyntacticPredicate) {
                //println("if (0==inputState.guessing)");
                //println("{");
                println("If ( inputState.guessing = 0 ) Then");

                tabadd();
            }

            ActionTransInfo tInfo = new ActionTransInfo();


            String actionStr = processActionForSpecialSymbols(action.actionText,
                action.getLine(), currentRule, tInfo);
            if (actionStr !=null && actionStr.toLowerCase().indexOf("if env")>-1) {
                int stp;
                stp=22;
            }
            if (tInfo.refRuleRoot != null) {
                // Somebody referenced "#rule", make sure translated var is valid
                // assignment to #rule is left as a ref also, meaning that assignments
                // with no other refs like "#rule = foo();" still forces this code to be
                // generated (unnecessarily).
                // // println(tInfo.refRuleRoot + " = ("+labeledElementASTType+")currentAST.root;");
                println(tInfo.refRuleRoot + " = DirectCast( currentAST.root, " +
                    labeledElementASTType + ")");
            }

            // dump the translated action
            printAction(actionStr);

            if (tInfo.assignToRoot) {
                // Somebody did a "#rule=", reset internal currentAST.root
                // // println("currentAST.root = "+tInfo.refRuleRoot+";");
                println("currentAST.root = " + tInfo.refRuleRoot);
                // reset the child pointer too to be last sibling in sibling list
                // // println("if ( (null != "+tInfo.refRuleRoot+") && (null != "+tInfo.refRuleRoot+".getFirstChild()) )");
                println("If ( (Not " + tInfo.refRuleRoot +
                    " Is Nothing) AndAlso ( Not " + tInfo.refRuleRoot +
                    ".getFirstChild() Is Nothing )) Then");

                tabadd();
                // // println("currentAST.child = "+tInfo.refRuleRoot+".getFirstChild();");
                println("currentAST.child = " + tInfo.refRuleRoot +
                    ".getFirstChild()");
                 tabsub();
                // // println("else");
                println("Else");

                tabadd();
                // // println("currentAST.child = "+tInfo.refRuleRoot+";");
                println("currentAST.child = " + tInfo.refRuleRoot);
                 tabsub();
                println("End If");
                // // println("currentAST.advanceChildToEnd();");
                println("currentAST.advanceChildToEnd()");
            }

            if (grammar.hasSyntacticPredicate) {
                 tabsub();
                // // println("}");
                println("End If");
            }
        }
    }

    /** Generate code for the given grammar element.
     * @param blk The "x|y|z|..." block to generate
     */
    public void gen(AlternativeBlock blk) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("gen(" + blk + ")");
            // // println("{");
            // output a empty line instead
        println("");

        // // // // tabadd();

        genBlockPreamble(blk);
        genBlockInitAction(blk);

        // Tell AST generation to build subrule result
        String saveCurrentASTResult = currentASTResult;
        if (blk.getLabel() != null) {
            currentASTResult = blk.getLabel();
        }

        boolean ok = grammar.theLLkAnalyzer.deterministic(blk);

        VBBlockFinishingInfo howToFinish = genCommonBlock(blk, true);
        genBlockFinish(howToFinish, throwNoViable);

        // // // // tabsub();
        // // println("}");
        // output a empty line instead
        println("");
        // Restore previous AST generation
        currentASTResult = saveCurrentASTResult;
    }

    /** Generate code for the given grammar element.
     * @param blk The block-end element to generate.  Block-end
     * elements are synthesized by the grammar parser to represent
     * the end of a block.
     */
    public void gen(BlockEndElement end) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genRuleEnd(" + end + ")");
    }

    /** Generate code for the given grammar element.
     * @param blk The character literal reference to generate
     */
    public void gen(CharLiteralElement atom) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genChar(" + atom + ")");

        if (atom.getLabel() != null) {
            // // println(atom.getLabel() + " = " + lt1Value + ";");
            // !?!
            println(atom.getLabel() + " = " + lt1Value);
        }

        boolean oldsaveText = saveText;
        saveText = saveText &&
            atom.getAutoGenType() == GrammarElement.AUTO_GEN_NONE;
        genMatch(atom);
        saveText = oldsaveText;
    }

    /** Generate code for the given grammar element.
     * @param blk The character-range reference to generate
     */
    public void gen(CharRangeElement r) {
        if (r.getLabel() != null && syntacticPredLevel == 0) {
            // // println(r.getLabel() + " = " + lt1Value + ";");
            // !?!
            println(r.getLabel() + " = " + lt1Value);
        }
        boolean flag = (grammar instanceof LexerGrammar &&
            (!saveText ||
            (r.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)));
        if (flag)
            println("_saveIndex = text.Length;");

            // // println("matchRange(" + OctalToUnicode(r.beginText) + "," + OctalToUnicode(r.endText) + ");");
        println("matchRange(" + convertCharLiteralPresentationToVB(OctalToUnicode(r.beginText)) + "," +
            convertCharLiteralPresentationToVB(OctalToUnicode(r.endText)) + ")");
        if (flag)

            // // println("text.Length = _saveIndex;");
            println("text.Length = _saveIndex");
    }

    /** Generate the lexer VB file */
    public void gen(LexerGrammar g) throws IOException {
        // If debugging, create a new sempred vector for this grammar
        if (g.debuggingOutput)
            semPreds = new Vector();

        setGrammar(g);
        if (! (grammar instanceof LexerGrammar)) {
            antlrTool.panic("Internal error generating lexer");
        }
        genBody(g);
    }

    /** Generate code for the given grammar element.
     * @param blk The (...)+ block to generate
     */
    public void gen(OneOrMoreBlock blk) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("gen+(" + blk + ")");
        String label;
        String cnt;
        // // println("{ // ( ... )+");
        println("\n' ' // ( ... )+");
        genBlockPreamble(blk);
        if (blk.getLabel() != null) {
            cnt = "_cnt_" + blk.getLabel();
        }
        else {
            cnt = "_cnt" + blk.ID;
        }
        // // println("int "+cnt+"=0;");
        println("Dim " + cnt + " As Integer = 0");
        if (blk.getLabel() != null) {
            label = blk.getLabel();
        }
        else {
            label = "_loop" + blk.ID;
        }

        // // println("for (;;)");
        println("While (True)");
        // // println("{");
        tabadd();
        // generate the init action for ()+ ()* inside the loop
        // this allows us to do usefull EOF checking...
        genBlockInitAction(blk);

        // Tell AST generation to build subrule result
        String saveCurrentASTResult = currentASTResult;
        if (blk.getLabel() != null) {
            currentASTResult = blk.getLabel();
        }

        boolean ok = grammar.theLLkAnalyzer.deterministic(blk);

        // generate exit test if greedy set to false
        // and an alt is ambiguous with exit branch
        // or when lookahead derived purely from end-of-file
        // Lookahead analysis stops when end-of-file is hit,
        // returning set {epsilon}.  Since {epsilon} is not
        // ambig with any real tokens, no error is reported
        // by deterministic() routines and we have to check
        // for the case where the lookahead depth didn't get
        // set to NONDETERMINISTIC (this only happens when the
        // FOLLOW contains real atoms + epsilon).
        boolean generateNonGreedyExitPath = false;
        int nonGreedyExitDepth = grammar.maxk;

        if (!blk.greedy &&
            blk.exitLookaheadDepth <= grammar.maxk &&
            blk.exitCache[blk.exitLookaheadDepth].containsEpsilon()) {
            generateNonGreedyExitPath = true;
            nonGreedyExitDepth = blk.exitLookaheadDepth;
        }
        else if (!blk.greedy &&
            blk.exitLookaheadDepth == LLkGrammarAnalyzer.NONDETERMINISTIC) {
            generateNonGreedyExitPath = true;
        }

        // generate exit test if greedy set to false
        // and an alt is ambiguous with exit branch
        if (generateNonGreedyExitPath) {
            if (DEBUG_CODE_GENERATOR) {
                System.out.println("nongreedy (...)+ loop; exit depth is " +
                    blk.exitLookaheadDepth);
            }
            String predictExit =
                getLookaheadTestExpression(blk.exitCache,
                nonGreedyExitDepth);
            // // println("// nongreedy exit test");
            println("' ' nongreedy exit test");
            // // println("if (("+cnt+" >= 1) && "+predictExit+") goto "+label+"_breakloop;");
            println("If ((" + cnt + " >= 1) AndAlso " + predictExit + ") Then");

            tabadd();
            println("GoTo " + label + "_breakloop");
             tabsub();
            println("End If");
        }

        VBBlockFinishingInfo howToFinish = genCommonBlock(blk, false);
        genBlockFinish(
            howToFinish,
            // // "if ("+cnt+" >= 1) { goto "+label+"_breakloop; } else { " + throwNoViable + "; }"
            // !?!
            "if (" + cnt + " >= 1) Then " + newline + tabstr +
            "\tGoTo " + label + "_breakloop " + newline+tabstr+
            "Else " + newline+tabstr+ "\t" +
            throwNoViable + newline+ tabstr+
            "End If"
            );

        // // println(cnt+"++;");
        println(cnt + " += 1");
         tabsub();
        // // println("}");
        // !?!  end of while loop
        println("End While");
        _print(label + "_breakloop:");
        // // println(";");
        println("");
        // // println("}    // ( ... )+");
        println("' ' // end of ( ... )+");

        // Restore previous AST generation
        currentASTResult = saveCurrentASTResult;
    }

    /** Generate the parser VB file */
    public void gen(ParserGrammar g) throws IOException {

        // if debugging, set up a new vector to keep track of sempred
        //   strings for this grammar
        if (g.debuggingOutput)
            semPreds = new Vector();

        setGrammar(g);
        if (! (grammar instanceof ParserGrammar)) {
            antlrTool.panic("Internal error generating parser");
        }
        genBody(g);
    }

    /** Generate code for the given grammar element.
     * @param blk The rule-reference to generate
     */
    public void gen(RuleRefElement rr) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genRR(" + rr + ")");
        RuleSymbol rs = (RuleSymbol) grammar.getSymbol(rr.targetRule);
        if (rs == null || !rs.isDefined()) {
            // Is this redundant???
            antlrTool.error("Rule '" + rr.targetRule + "' is not defined",
                grammar.getFilename(), rr.getLine(), rr.getColumn());
            return;
        }
        if (! (rs instanceof RuleSymbol)) {
            // Is this redundant???
            antlrTool.error("'" + rr.targetRule +
                "' does not name a grammar rule",
                grammar.getFilename(), rr.getLine(), rr.getColumn());
            return;
        }

        genErrorTryForElement(rr);

        // AST value for labeled rule refs in tree walker.
        // This is not AST construction;  it is just the input tree node value.
        if (grammar instanceof TreeWalkerGrammar &&
            rr.getLabel() != null &&
            syntacticPredLevel == 0) {
            // // println(rr.getLabel() + " = _t==ASTNULL ? null : "+lt1Value+";");
            println(rr.getLabel() + " = IIf( _t Is ASTNULL,  Nothing, " +
                lt1Value + " )");
        }

        // if in lexer and ! on rule ref or alt or rule, save buffer index to kill later
        if (grammar instanceof LexerGrammar &&
            (!saveText || rr.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
            declareSaveIndexVariableIfNeeded();
            // // println("_saveIndex = text.Length;");
            println("_saveIndex = text.Length");
        }

        // Process return value assignment if any
        printTabs();
        if (rr.idAssign != null) {
            // Warn if the rule has no return type
            if (rs.block.returnAction == null) {
                antlrTool.warning("Rule '" + rr.targetRule +
                    "' has no return type", grammar.getFilename(),
                    rr.getLine(), rr.getColumn());
            }
            _print(rr.idAssign + "=");
        }
        else {
            // Warn about return value if any, but not inside syntactic predicate
            if (! (grammar instanceof LexerGrammar) && syntacticPredLevel == 0 &&
                rs.block.returnAction != null) {
                antlrTool.warning("Rule '" + rr.targetRule +
                    "' returns a value", grammar.getFilename(),
                    rr.getLine(), rr.getColumn());
            }
        }

        // Call the rule
        GenRuleInvocation(rr);

        // if in lexer and ! on element or alt or rule, save buffer index to kill later
        if (grammar instanceof LexerGrammar &&
            (!saveText || rr.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
            declareSaveIndexVariableIfNeeded();
            // // println("text.Length = _saveIndex;");
            println("text.Length = _saveIndex");
        }

        // if not in a syntactic predicate
        if (syntacticPredLevel == 0) {
            boolean doNoGuessTest = (
                grammar.hasSyntacticPredicate &&
                (
                grammar.buildAST && rr.getLabel() != null ||
                (genAST && rr.getAutoGenType() == GrammarElement.AUTO_GEN_NONE)
                )
                );
            if (doNoGuessTest) {
                // // println("if (0 == inputState.guessing)");
                // // println("{");
                println("If (inputState.guessing = 0) Then");

                tabadd();
            }

            if (grammar.buildAST && rr.getLabel() != null) {
                // always gen variable for rule return on labeled rules
                // // println(rr.getLabel() + "_AST = ("+labeledElementASTType+")returnAST;");
                println(rr.getLabel() + "_AST = DirectCast( returnAST, " +
                    labeledElementASTType + ")");
            }
            if (genAST) {
                switch (rr.getAutoGenType()) {
                    case GrammarElement.AUTO_GEN_NONE:
                        if (usingCustomAST)

                            // // println("astFactory.addASTChild(currentAST, (AST)returnAST);");
                            println(
                                "astFactory.addASTChild(currentAST, DirectCast(returnAST, AST))");
                        else

                            // // println("astFactory.addASTChild(currentAST, returnAST);");
                            println(
                                "astFactory.addASTChild(currentAST, returnAST)");
                        break;
                    case GrammarElement.AUTO_GEN_CARET:
                        antlrTool.error(
                            "Internal: encountered ^ after rule reference");
                        break;
                    default:
                        break;
                }
            }

            // if a lexer and labeled, Token label defined at rule level, just set it here
            if (grammar instanceof LexerGrammar && rr.getLabel() != null) {
                // // println(rr.getLabel()+" = returnToken_;");
                println(rr.getLabel() + " = returnToken_");
            }

            if (doNoGuessTest) {
                 tabsub();
                // // println("}");
                println(
                    "End If ' ' // end of If (inputState.guessing = 0) Then"); // end of println("If (inputState.guessing = 0) Then");
            }
        }
        genErrorCatchForElement(rr);
    }

    /** Generate code for the given grammar element.
     * @param blk The string-literal reference to generate
     */
    public void gen(StringLiteralElement atom) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genString(" + atom + ")");

            // Variable declarations for labeled elements
        if (atom.getLabel() != null && syntacticPredLevel == 0) {
            // // println(atom.getLabel() + " = " + lt1Value + ";");
            println(atom.getLabel() + " = " + lt1Value);
        }

        // AST
        genElementAST(atom);

        // is there a bang on the literal?
        boolean oldsaveText = saveText;
        saveText = saveText &&
            atom.getAutoGenType() == GrammarElement.AUTO_GEN_NONE;

        // matching
        genMatch(atom);

        saveText = oldsaveText;

        // tack on tree cursor motion if doing a tree walker
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("_t = _t.getNextSibling();");
            println("_t = _t.getNextSibling()");
        }
    }

    /** Generate code for the given grammar element.
     * @param blk The token-range reference to generate
     */
    public void gen(TokenRangeElement r) {
        genErrorTryForElement(r);
        if (r.getLabel() != null && syntacticPredLevel == 0) {
            // // println(r.getLabel() + " = " + lt1Value + ";");
            println(r.getLabel() + " = " + lt1Value);
        }

        // AST
        genElementAST(r);

        // match
        // // println("matchRange("+OctalToUnicode(r.beginText)+","+OctalToUnicode(r.endText)+");");
        println("matchRange(" + convertCharLiteralPresentationToVB(OctalToUnicode(r.beginText)) + "," +
            convertCharLiteralPresentationToVB(OctalToUnicode(r.endText)) + ")");
        genErrorCatchForElement(r);
    }

    /** Generate code for the given grammar element.
     * @param blk The token-reference to generate
     */
    public void gen(TokenRefElement atom) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genTokenRef(" + atom + ")");
        if (grammar instanceof LexerGrammar) {
            antlrTool.panic("Token reference found in lexer");
        }
        genErrorTryForElement(atom);
        // Assign Token value to token label variable
        if (atom.getLabel() != null && syntacticPredLevel == 0) {
            // // println(atom.getLabel() + " = " + lt1Value + ";");
            println(atom.getLabel() + " = " + lt1Value);
        }

        // AST
        genElementAST(atom);
        // matching
        genMatch(atom);
        genErrorCatchForElement(atom);

        // tack on tree cursor motion if doing a tree walker
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("_t = _t.getNextSibling();");
            println("_t = _t.getNextSibling()");
        }
    }

    public void gen(TreeElement t) {
        // save AST cursor
        // // println("AST __t" + t.ID + " = _t;");
        println("Dim __t" + t.ID + " As AST = _t");

        // If there is a label on the root, then assign that to the variable
        if (t.root.getLabel() != null) {
            // // println(t.root.getLabel() + " = (ASTNULL == _t) ? null : ("+labeledElementASTType +")_t;");
            println(t.root.getLabel() +
                " = IIf( _t Is ASTNULL,  Nothing, DirectCast( _t, " +
                labeledElementASTType + "))");
        }

        // check for invalid modifiers ! and ^ on tree element roots
        if (t.root.getAutoGenType() == GrammarElement.AUTO_GEN_BANG) {
            antlrTool.error("Suffixing a root node with '!' is not implemented",
                grammar.getFilename(), t.getLine(), t.getColumn());
            t.root.setAutoGenType(GrammarElement.AUTO_GEN_NONE);
        }
        if (t.root.getAutoGenType() == GrammarElement.AUTO_GEN_CARET) {
            antlrTool.warning(
                "Suffixing a root node with '^' is redundant; already a root",
                grammar.getFilename(), t.getLine(), t.getColumn());
            t.root.setAutoGenType(GrammarElement.AUTO_GEN_NONE);
        }

        // Generate AST variables
        genElementAST(t.root);
        if (grammar.buildAST) {
            // Save the AST construction state
            // // println("ASTPair __currentAST" + t.ID + " = currentAST.copy();");
            println("Dim __currentAST" + t.ID + " As ASTPair = currentAST.copy()");
            // Make the next item added a child of the TreeElement root
            // // println("currentAST.root = currentAST.child;");
            println("currentAST.child = Nothing");
            // // println("currentAST.root = currentAST.child;");
            println("currentAST.child = Nothing");
        }

        // match root
        if (t.root instanceof WildcardElement) {
            // // println("if (null == _t) throw new MismatchedTokenException();");
            println(
                "if (_t Is Nothing ) Then throw new MismatchedTokenException()");
        }
        else {
            genMatch(t.root);
        }
        // move to list of children
        // // println("_t = _t.getFirstChild();");
        println("_t = _t.getFirstChild()");

        // walk list of children, generating code for each
        for (int i = 0; i < t.getAlternatives().size(); i++) {
            Alternative a = t.getAlternativeAt(i);
            AlternativeElement e = a.head;
            while (e != null) {
                e.generate();
                e = e.next;
            }
        }

        if (grammar.buildAST) {
            // restore the AST construction state to that just after the
            // tree root was added
            // // println("currentAST = __currentAST" + t.ID + ";");
            println("currentAST = __currentAST" + t.ID);
        }
        // restore AST cursor
        // // println("_t = __t" + t.ID + ";");
        println("_t = __t" + t.ID);
        // move cursor to sibling of tree just parsed
        // // println("_t = _t.getNextSibling();");
        println("_t = _t.getNextSibling()");

    }

    /** Generate the tree-parser VB file */
    public void gen(TreeWalkerGrammar g) throws IOException {
        // SAS: debugging stuff removed for now...
        setGrammar(g);
        if (! (grammar instanceof TreeWalkerGrammar)) {
            antlrTool.panic("Internal error generating tree-walker");
        }
        genBody(g);
    }

    /** Generate code for the given grammar element.
     * @param wc The wildcard element to generate
     */
    public void gen(WildcardElement wc) {
        // Variable assignment for labeled elements
        if (wc.getLabel() != null && syntacticPredLevel == 0) {
            // // println(wc.getLabel() + " = " + lt1Value + ";");
            println(wc.getLabel() + " = " + lt1Value);
        }

        // AST
        genElementAST(wc);
        // Match anything but EOF
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("if (null == _t) throw new MismatchedTokenException();");
            println("if (_t Is Nothing ) Then throw new MismatchedTokenException()");
        }
        else if (grammar instanceof LexerGrammar) {
            if (grammar instanceof LexerGrammar &&
                (!saveText ||
                wc.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
                declareSaveIndexVariableIfNeeded();
                // // println("_saveIndex = text.Length;");
                println("_saveIndex = text.Length");
            }
            // // println("matchNot(EOF/*_CHAR*/);");
            println("matchNot(EOF/*_CHAR*/)");
            if (grammar instanceof LexerGrammar &&
                (!saveText ||
                wc.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
                declareSaveIndexVariableIfNeeded();
                // // println("text.Length = _saveIndex;"); // kill text atom put in buffer
                println("text.Length = _saveIndex");
            }
        }
        else {
            // // println("matchNot(" + getValueString(Token.EOF_TYPE) + ");");
            println("matchNot(" + getValueString(Token.EOF_TYPE) + ")");
        }

        // tack on tree cursor motion if doing a tree walker
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("_t = _t.getNextSibling();");
            println("_t = _t.getNextSibling()");
        }
    }

    /** Generate code for the given grammar element.
     * @param blk The (...)* block to generate
     */
    public void gen(ZeroOrMoreBlock blk) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("gen*(" + blk + ")");
            // // println("{    // ( ... )*");
        println("' ' // ( ... )*");
        // //
        genBlockPreamble(blk);
        String label;
        if (blk.getLabel() != null) {
            label = blk.getLabel();
        }
        else {
            label = "_loop" + blk.ID;
        }
        // // println("for (;;)");
        println("While (True)"); // printing block=1
        // // println("{");

        tabadd();
        // generate the init action for ()+ ()* inside the loop
        // this allows us to do usefull EOF checking...
        genBlockInitAction(blk);

        // Tell AST generation to build subrule result
        String saveCurrentASTResult = currentASTResult;
        if (blk.getLabel() != null) {
            currentASTResult = blk.getLabel();
        }

        boolean ok = grammar.theLLkAnalyzer.deterministic(blk);

        // generate exit test if greedy set to false
        // and an alt is ambiguous with exit branch
        // or when lookahead derived purely from end-of-file
        // Lookahead analysis stops when end-of-file is hit,
        // returning set {epsilon}.  Since {epsilon} is not
        // ambig with any real tokens, no error is reported
        // by deterministic() routines and we have to check
        // for the case where the lookahead depth didn't get
        // set to NONDETERMINISTIC (this only happens when the
        // FOLLOW contains real atoms + epsilon).
        boolean generateNonGreedyExitPath = false;
        int nonGreedyExitDepth = grammar.maxk;

        if (!blk.greedy &&
            blk.exitLookaheadDepth <= grammar.maxk &&
            blk.exitCache[blk.exitLookaheadDepth].containsEpsilon()) {
            generateNonGreedyExitPath = true;
            nonGreedyExitDepth = blk.exitLookaheadDepth;
        }
        else if (!blk.greedy &&
            blk.exitLookaheadDepth == LLkGrammarAnalyzer.NONDETERMINISTIC) {
            generateNonGreedyExitPath = true;
        }
        if (generateNonGreedyExitPath) {
            if (DEBUG_CODE_GENERATOR) {
                System.out.println("nongreedy (...)* loop; exit depth is " +
                    blk.exitLookaheadDepth);
            }
            String predictExit =
                getLookaheadTestExpression(blk.exitCache,
                nonGreedyExitDepth);
            // // println("// nongreedy exit test");
            println("' ' nongreedy exit test");
            // // println("if ("+predictExit+") goto "+label+"_breakloop;");
            println("If (" + predictExit + ") Then GoTo " + label +
                "_breakloop");
        }

        VBBlockFinishingInfo howToFinish = genCommonBlock(blk, false);
        genBlockFinish(howToFinish, "GoTo " + label + "_breakloop");

         tabsub();
        // // println("}");
        println("End While"); // printing block = 1 - 1 = 0
        _print(label + "_breakloop:");
        // // println(";");
        println("");
        // //  tabsub();
        // // println("}    // ( ... )*");
        println("' ' // end of ( ... )*");

        // Restore previous AST generation
        currentASTResult = saveCurrentASTResult;
    }

    /** Generate an alternative.
     * @param alt  The alternative to generate
     * @param blk The block to which the alternative belongs
     */
    protected void genAlt(Alternative alt, AlternativeBlock blk) {
        // Save the AST generation state, and set it to that of the alt
        boolean savegenAST = genAST;
        genAST = genAST && alt.getAutoGen();

        boolean oldsaveTest = saveText;
        saveText = saveText && alt.getAutoGen();

        // Reset the variable name map for the alternative
        Hashtable saveMap = treeVariableMap;
        treeVariableMap = new Hashtable();

        // Generate try block around the alt for  error handling
        if (alt.exceptionSpec != null) {
            // // println("try        // for error handling");
            println("Try        ' ' // for error handling"); // printing block = 0 + 1 = 1
            // // println("{");
            tabadd();
        }

        AlternativeElement elem = alt.head;
        while (! (elem instanceof BlockEndElement)) {
            elem.generate(); // alt can begin with anything. Ask target to gen.
            elem = elem.next;
        }

        if (genAST) {
            if (blk instanceof RuleBlock) {
                // Set the AST return value for the rule
                RuleBlock rblk = (RuleBlock) blk;
                if (usingCustomAST) {
                    // // println(rblk.getRuleName() + "_AST = ("+labeledElementASTType+")currentAST.root;");
                    println(rblk.getRuleName() +
                        "_AST = DirectCast(currentAST.root, " +
                        labeledElementASTType + ")");
                }
                else {
                    // // println(rblk.getRuleName() + "_AST = currentAST.root;");
                    println(rblk.getRuleName() + "_AST = currentAST.root");
                }
            }
            else if (blk.getLabel() != null) {
                // ### future: also set AST value for labeled subrules.
                // println(blk.getLabel() + "_AST = ("+labeledElementASTType+")currentAST.root;");
                antlrTool.warning("Labeled subrules not yet supported",
                    grammar.getFilename(), blk.getLine(),
                    blk.getColumn());
            }
        }

        if (alt.exceptionSpec != null) {
            // close try block
             tabsub();
            // // println("}");
            genErrorHandler(alt.exceptionSpec);
            // added: 'End Try'
            println("End Try"); // after a Catch clause, need an extra End Try
        }

        genAST = savegenAST;
        saveText = oldsaveTest;

        treeVariableMap = saveMap;
    }

    /** Generate all the bitsets to be used in the parser or lexer
     * Generate the raw bitset data like "long _tokenSet1_data[] = {...};"
     * and the BitSet object declarations like "BitSet _tokenSet1 = new BitSet(_tokenSet1_data);"
     * Note that most languages do not support object initialization inside a
     * class definition, so other code-generators may have to separate the
     * bitset declarations from the initializations (e.g., put the initializations
     * in the generated constructor instead).
     * @param bitsetList The list of bitsets to generate.
     * @param maxVocabulary Ensure that each generated bitset can contain at least this value.
     */
    protected void genBitsets(Vector bitsetList, int maxVocabulary) {
        println("");
        for (int i = 0; i < bitsetList.size(); i++) {
            BitSet p = (BitSet) bitsetList.elementAt(i);
            // Ensure that generated BitSet is large enough for vocabulary
            p.growToInclude(maxVocabulary);
            genBitSet(p, i);
        }
    }

    /** Do something simple like:
     *  private static final long[] mk_tokenSet_0() {
     *    long[] data = { -2305839160922996736L, 63L, 16777216L, 0L, 0L, 0L };
     *    return data;
     *  }
     *  public static final BitSet _tokenSet_0 = new BitSet(mk_tokenSet_0());
     *
     *  Or, for large bitsets, optimize init so ranges are collapsed into loops.
     *  This is most useful for lexers using unicode.
     */
    private void genBitSet(BitSet p, int id) {
        // initialization data
        // // println("private static long[] mk_" + getBitsetName(id) + "()");
        println("private Shared Function mk_" + getBitsetName(id) +
            "() As Long()");
        // // println("{");
        // start of the function definition block

        tabadd();
        int n = p.lengthInLongWords();
        if (n < BITSET_OPTIMIZE_INIT_THRESHOLD) {
            // // println("long[] data = { " + p.toStringOfWords() + "};");
            println("Dim data() as Long = { " + p.toStringOfWords() + "}");
        }
        else {
            // will init manually, allocate space then set values
            // // println("long[] data = new long["+n+"];");
            println("Dim data() as Long: Redim data(" +n + "-1)");
            boolean isIdef = false;
            long[] elems = p.toPackedArray();
            for (int i = 0; i < elems.length; ) {
                if ( (i + 1) == elems.length || elems[i] != elems[i + 1]) {
                    // last number or no run of numbers, just dump assignment
                    // // println("data["+i+"]="+elems[i]+"L;");
                    println("data(" + i + ")=" + elems[i] + "L");
                    i++;
                }
                else {
                    // scan to find end of run
                    int j;
                    for (j = i + 1; j < elems.length && elems[j] == elems[i]; j++) {
                        ;
                    }
                    // j-1 is last member of run
                    // // println("for (int i = "+i+"; i<="+(j-1)+"; i++) { data[i]="+ elems[i]+"L; }");
                    if (!isIdef) {
                        println("Dim i as Integer");
                        isIdef = true;
                    }
                    println("For i = " + i + "To " + (j - 1));

                    tabadd();
                    println("data(i)=" + elems[i] + "L");
                     tabsub();
                    println("Next");
                    i = j;
                }
            }
        }

        // // println("return data;");
        println("return data");
         tabsub(); // end of the function definition block
        println("End Function");
        // the followed by an object created by using the previously defined function
        // BitSet object
        // // println("public static readonly BitSet " + getBitsetName(id) + " = new BitSet(" + "mk_" + getBitsetName(id) + "()" + ");");
        println("Public Shared ReadOnly " + getBitsetName(id) +
            " As BitSet = New BitSet( mk_" + getBitsetName(id) + "()" + ")");
    }

    /** Given the index of a bitset in the bitset list, generate a unique name.
     * Specific code-generators may want to override this
     * if the language does not allow '_' or numerals in identifiers.
     * @param index  The index of the bitset in the bitset list.
     */
    protected String getBitsetName(int index) {
        return "tokenSet_" + index + "_";
    }

    /** Generate the finish of a block, using a combination of the info
     * returned from genCommonBlock() and the action to perform when
     * no alts were taken
     * @param howToFinish The return of genCommonBlock()
     * @param noViableAction What to generate when no alt is taken
     */
    private void genBlockFinish(VBBlockFinishingInfo howToFinish,
        String noViableAction) {

        if (howToFinish.needAnErrorClause &&
            (howToFinish.generatedAnIf || howToFinish.generatedSwitch)) {
            if (howToFinish.generatedAnIf) {
                // // println("else");
                println("Else"); // // for prepare the exception part
                // // // // println( newline + tabstr + "Else"); // !?! !?! !?! ??????
                // // println("{");
            }
            else {
                // // println("{");
            }

            tabadd();

            println(noViableAction);
            tabsub();
            // // println("}");
            // // println("End If"); // guessing, must be If statement -- should not be printed
            if (howToFinish.generatedAnIf ) {
                println("End If"); // // This is added // !?! correct!
            }
        }
        // // tabsub();
        if (howToFinish.postscript != null) {
            if (howToFinish.needAnErrorClause && howToFinish.generatedSwitch &&
                !howToFinish.generatedAnIf && noViableAction != null) {
                // Check to make sure that noViableAction is only a throw statement
                // // if (noViableAction.indexOf("throw") == 0 || noViableAction.indexOf("goto") == 0) {
                if (noViableAction.toLowerCase().indexOf("throw") == 0 ||
                    noViableAction.toLowerCase().indexOf("goto") == 0) {
                    // Remove the break statement since it isn't reachable with a throw exception
                    // x?x x?x
                    int endOfBreak = howToFinish.postscript.indexOf("break;") +
                        6;
                    // // String newPostScript = howToFinish.postscript.substring(endOfBreak);
                    String newPostScript = howToFinish.postscript;
                    println(newPostScript);
                }
                else {
                    println(howToFinish.postscript);
                }
            }
            else {
                println(howToFinish.postscript);
            }
        }
    }

    /** Generate the init action for a block, which may be a RuleBlock or a
     * plain AlternativeBLock.
     * @blk The block for which the preamble is to be generated.
     */
    protected void genBlockInitAction(AlternativeBlock blk) {
        // dump out init action
        if (blk.initAction != null) {
            // // printAction(processActionForSpecialSymbols(blk.initAction, blk.getLine(), currentRule, null));
            println(processActionForSpecialSymbols(blk.initAction, blk.getLine(), currentRule, null));
        }
    }


    /*
     * Overrides the parent class's _printAction method
     */
    protected void _printAction(String s) {
        if (s == null) {
            return;
        }

        // Skip leading newlines, tabs and spaces
        int start = 0;
        int end = s.length() - 1;
        // // while (start < s.length() && Character.isSpaceChar(s.charAt(start))) {
        // //     start++;
        // // }
        // Skip leading newlines, tabs and spaces
        // //
        // // while (end > start && Character.isSpaceChar(s.charAt(end))) {
        // //     end--;
        // // }

        s = s.trim();
        end = s.length()-1;
        char c = 0;
        for (int i = start; i <= end; ) {
            c = s.charAt(i);
            i++;
            boolean newline = false;
            switch (c) {
                case '\n':
                    newline = true;
                    break;
                case '\r':
                    if (i <= end && s.charAt(i) == '\n') {
                        i++;
                    }
                    newline = true;
                    break;
                default:
                    currentOutput.print(c);
                    break;
            }
            if (newline) {
                currentOutput.println();
                // // printTabs(); // !?!
                // Absorb leading whitespace
                while (i <= end && Character.isSpaceChar(s.charAt(i))) {
                    i++;
                }
                newline = false;
            }
        }
        currentOutput.println();
} // end of function _printAction(String s)


    /** Generate the header for a block, which may be a RuleBlock or a
     * plain AlternativeBLock.  This generates any variable declarations
     * and syntactic-predicate-testing variables.
     * @blk The block for which the preamble is to be generated.
     */
    protected void genBlockPreamble(AlternativeBlock blk) {
        // define labels for rule blocks.
        if (blk instanceof RuleBlock) {
            RuleBlock rblk = (RuleBlock) blk;
            if (rblk.labeledElements != null) {
                for (int i = 0; i < rblk.labeledElements.size(); i++) {

                    AlternativeElement a = (AlternativeElement) rblk.
                        labeledElements.elementAt(i);
                    //System.out.println("looking at labeled element: "+a);
                    //Variables for labeled rule refs and
                    //subrules are different than variables for
                    //grammar atoms.  This test is a little tricky
                    //because we want to get all rule refs and ebnf,
                    //but not rule blocks or syntactic predicates
                    if (a instanceof RuleRefElement || a instanceof AlternativeBlock &&
                        ! (a instanceof RuleBlock) && ! (a instanceof SynPredBlock)) {
                        if (
                            ! (a instanceof RuleRefElement) &&
                            ( (AlternativeBlock) a).not &&
                            analyzer.subruleCanBeInverted( ( (AlternativeBlock)
                            a), grammar instanceof LexerGrammar)
                            ) {
                            // Special case for inverted subrules that
                            // will be inlined.  Treat these like
                            // token or char literal references
                            // // println(labeledElementType + " " + a.getLabel() + " = " + labeledElementInit + ";");
                            println("Dim " + a.getLabel() + " As " + labeledElementType + " = " + labeledElementInit);
                            if (grammar.buildAST) {
                                genASTDeclaration(a);
                            }
                        }
                        else {
                            if (grammar.buildAST) {
                                // Always gen AST variables for
                                // labeled elements, even if the
                                // element itself is marked with !
                                genASTDeclaration(a);
                            }
                            if (grammar instanceof LexerGrammar) {
                                // // println("Token "+a.getLabel()+" = null;");
                                println("Dim " + a.getLabel() + " As Token = Nothing");
                            }
                            if (grammar instanceof TreeWalkerGrammar) {
                                // always generate rule-ref variables
                                // for tree walker
                                // // println(labeledElementType + " " + a.getLabel() + " = " + labeledElementInit + ";");
                                println("Dim " + a.getLabel() + " As " + labeledElementType + " = " + labeledElementInit);
                            }
                        }
                    }
                    else {
                        // It is a token or literal reference.  Generate the
                        // correct variable type for this grammar
                        // // println(labeledElementType + " " + a.getLabel() + " = " + labeledElementInit + ";");
                        println("Dim " + a.getLabel() + " As " + labeledElementType + " = " + labeledElementInit);
                        // In addition, generate *_AST variables if building ASTs
                        if (grammar.buildAST) {
                            //println(labeledElementASTType+" " + a.getLabel() + "_AST = null;");
                            if (a instanceof GrammarAtom &&
                                ( (GrammarAtom) a).getASTNodeType() != null) {
                                GrammarAtom ga = (GrammarAtom) a;
                                genASTDeclaration(a, ga.getASTNodeType());
                            }
                            else {
                                genASTDeclaration(a);
                            }
                        }
                    }
                }
            }
        }
    }
    // generate lexer body
    public void genBody(LexerGrammar g) throws IOException {
        // SAS: moved output creation to method so a subclass can change
        //      how the output is generated (for VAJ interface)
        setupOutput(grammar.getClassName());

        genAST = false; // no way to gen trees.
        saveText = true; // save consumed characters.

        // // tabs = 0;
        tabclear();
        // Generate header common to all vb output files
        genHeader();
        // Do not use printAction because we assume tabs==0
        println(behavior.getHeaderAction(""));

        // Generate the vb namespace declaration (if specified)
        /* For VB, namespace declaration is after Imports statements
        if (nameSpace != null)
            nameSpace.emitDeclarations(currentOutput);

        tabadd();
        */

        // Generate header specific to lexer VB file
        // println("import java.io.FileInputStream;");
        // // println("// Generate header specific to lexer vb file");
        println("' ' // Generate header specific to lexer VB file");
        // // println("using System;");
        println("Imports System");
        // // println("using Stream                          = System.IO.Stream;");
        println("Imports Stream                        = System.IO.Stream");
        // // println("using TextReader                      = System.IO.TextReader;");
        println("Imports TextReader                    = System.IO.TextReader");
        // // println("using Hashtable                       = System.Collections.Hashtable;");
        println("Imports Hashtable                     = System.Collections.Hashtable");
        // // println("using Comparer                        = System.Collections.Comparer;");
        println("Imports Comparer                      = System.Collections.Comparer");

        if (! (g.caseSensitiveLiterals)) {
            // // println("using CaseInsensitiveHashCodeProvider = System.Collections.CaseInsensitiveHashCodeProvider;");
            println("Imports CaseInsensitiveHashCodeProvider = System.Collections.CaseInsensitiveHashCodeProvider");
            // // println("using CaseInsensitiveComparer         = System.Collections.CaseInsensitiveComparer;");
            println("Imports CaseInsensitiveComparer         = System.Collections.CaseInsensitiveComparer");
        }
        println("");


        // // println("Imports  TokenStreamException            = antlr.TokenStreamException");
        // // println("Imports  TokenStreamIOException          = antlr.TokenStreamIOException");
        // // println("Imports  TokenStreamRecognitionException = antlr.TokenStreamRecognitionException");
        // // println("Imports  CharStreamException             = antlr.CharStreamException");
        // // println("Imports  CharStreamIOException           = antlr.CharStreamIOException");
        // // println("Imports  ANTLRException                  = antlr.ANTLRException");
        // // println("Imports  CharScanner                     = antlr.CharScanner");
        // // println("Imports  InputBuffer                     = antlr.InputBuffer");
        // // println("Imports  ByteBuffer                      = antlr.ByteBuffer");
        // // println("Imports  CharBuffer                      = antlr.CharBuffer");
        // // println("Imports  Token                           = antlr.Token");
        // // println("Imports  CommonToken                     = antlr.CommonToken");
        // // println("Imports  SemanticException               = antlr.SemanticException");
        // // println("Imports  RecognitionException            = antlr.RecognitionException");
        // // println("Imports  NoViableAltForCharException     = antlr.NoViableAltForCharException");
        // // println("Imports  MismatchedCharException         = antlr.MismatchedCharException");
        // // println("Imports  TokenStream                     = antlr.TokenStream");
        // // println("Imports  LexerSharedInputState           = antlr.LexerSharedInputState");
        // // println("Imports  BitSet                          = antlr.collections.impl.BitSet");
        println("Imports  antlr");
        println("Imports  antlr.collections");
        println("Imports  antlr.collections.impl");
        println("Imports  antlr.debug");
        println("Imports  antlr.debug.misc");
        // Generate user-defined lexer file preamble
        println(grammar.preambleAction.getText());

        if (nameSpace != null){
                  nameSpace.emitDeclarations(currentOutput);
              }
        tabadd();
        println("");



        // Generate lexer class definition
        String sup = null;
        if (grammar.superClass != null) {
            sup = grammar.superClass;
        }
        else {
            sup = "antlr." + grammar.getSuperClass();
        }

        // print javadoc comment if any
        if (grammar.comment != null) {
            _println(grammar.comment);
        }

        Token tprefix = (Token) grammar.options.get("classHeaderPrefix");
        if (tprefix == null) {
            // // print("public ");
            print("Public ");
        }
        else {
            // !?! This one seems taking off the double quotes arround a literal string
            String p = StringUtils.stripFrontBack(tprefix.getText(), "\"", "\"");
            if (p == null) {
                // // print("public ");
                print("Public ");
            }
            else {
                print(p + " ");
            }
        }

        // // print("class " + grammar.getClassName() + " : " + sup);
        println("Class " + grammar.getClassName());
        println("Inherits " + sup);
        String suffix="";

        Token tsuffix = (Token) grammar.options.get("classHeaderSuffix");
        if (tsuffix != null) {
            // // print("Implements TokenStream");
            suffix = StringUtils.stripFrontBack(tsuffix.getText(), "\"",
                "\"");
            if (suffix != null) {
                print("Implements " + suffix); // must be an interface name for VB
            }
        }
        println(""); // ending the interfaces
        // // println(" {"); Printing block =0 + 1

        tabadd();

        // Generate 'const' definitions for Token IDs
        genTokenDefinitions(grammar.tokenManager);

        // Generate user-defined lexer class members
        print(
            processActionForSpecialSymbols(grammar.classMemberAction.getText(),
            grammar.classMemberAction.getLine(),
            currentRule, null)
            );

        //
        // Generate the constructor from InputStream, which in turn
        // calls the ByteBuffer constructor
        //
        // // println("public " + grammar.getClassName() +"(Stream ins) : this(new ByteBuffer(ins))");
        println("Public Sub New (ins As Stream )");
        // // : this(new ByteBuffer(ins))"); // put this into the constructor
        // // println("{");

        tabadd();
        println("MyClass.New( New ByteBuffer(ins))");
        // // println("}");
         tabsub();
        println("End Sub"); // end of the constructor
        println(""); // an empty line

        //
        // Generate the constructor from Reader, which in turn
        // calls the CharBuffer constructor
        //
        // // println("public " + grammar.getClassName() + "(TextReader r) : this(new CharBuffer(r))");
        println("Public Sub New(r As TextReader)");
        // //  : this(new CharBuffer(r))");
        // // println("{");

        tabadd();
        println("MyClass.New( New CharBuffer(r))");
        // // println("}");
         tabsub();
        println("End Sub");
        println("");

        // // print("public " + grammar.getClassName() + "(InputBuffer ib)");
        println("Public Sub New(ib As InputBuffer)");
        // if debugging, wrap the input buffer in a debugger

        tabadd();
        if (grammar.debuggingOutput)

            // // println(" : this(new LexerSharedInputState(new antlr.debug.DebuggingInputBuffer(ib)))");
            println("MyClass.New(New LexerSharedInputState(New antlr.debug.DebuggingInputBuffer(ib)))");
        else
            println("MyClass.New(New LexerSharedInputState(ib))");
            // // println("{");
         tabsub();
        // // println("}");
        println("End Sub");
        println("");

        //
        // Generate the constructor from InputBuffer (char or byte)
        //
        // // println("public " + grammar.getClassName() + "(LexerSharedInputState state) : base(state)");
        println("Public Sub New(state As LexerSharedInputState)");
        // // : base(state)");

        tabadd();
        // // println("{");
        println("MyBase.New(state)");
        // // println("initialize();");
        println("initialize()");
         tabsub();
        // // println("}");
        println("End Sub");

        // ***** another function
        // Generate the initialize function
        // // println("private void initialize()");
        println("Private Sub initialize()");
        // // println("{");

        tabadd();

        // if debugging, set up array variables and call user-overridable
        //   debugging setup method
        if (grammar.debuggingOutput) {
            // // println("ruleNames  = _ruleNames;");
            println("ruleNames  = _ruleNames");
            // // println("semPredNames = _semPredNames;");
            println("semPredNames = _semPredNames");
            // // println("setupDebugging();");
            println("setupDebugging()");
        }

        // Generate the setting of various generated options.
        // These need to be before the literals since ANTLRHashString depends on
        // the casesensitive stuff.
        // // println("caseSensitiveLiterals = " + g.caseSensitiveLiterals + ";");
        println("caseSensitiveLiterals = " + g.caseSensitiveLiterals);
        // // println("setCaseSensitive(" + g.caseSensitive + ");");
        println("setCaseSensitive(" + g.caseSensitive + ")");

        // Generate the initialization of a hashtable
        // containing the string literals used in the lexer
        // The literals variable itself is in CharScanner
        if (g.caseSensitiveLiterals)
            // // println("literals = new Hashtable(100, (float) 0.4, null, Comparer.Default);");
            println("literals = new Hashtable(100, 0.4F, Nothing, Comparer.Default)");
        else
            // // println("literals = new Hashtable(100, (float) 0.4, CaseInsensitiveHashCodeProvider.Default, CaseInsensitiveComparer.Default);");
            println("literals = new Hashtable(100, 0.4F, CaseInsensitiveHashCodeProvider.Default, CaseInsensitiveComparer.Default)");
        Enumeration keys = grammar.tokenManager.getTokenSymbolKeys();
        while (keys.hasMoreElements()) {
            String key = (String) keys.nextElement();
            if (key.charAt(0) != '"') {
                continue;
            }
            TokenSymbol sym = grammar.tokenManager.getTokenSymbol(key);
            if (sym instanceof StringLiteralSymbol) {
                StringLiteralSymbol s = (StringLiteralSymbol) sym;
                // // println("literals.Add(" + s.getId() + ", " + s.getTokenType() + ");");
                println("literals.Add(" + s.getId() + ", " + s.getTokenType() + ")");
            }
        }

        Enumeration ids;
         tabsub();
        // // println("}");
        println("End Sub");

        // *** a field
        // generate the rule name array for debugging
        if (grammar.debuggingOutput) {

            // // println("private const string[] _ruleNames = {");
            println("Private Shared _ruleNames() As String = _ruleNames_Extract()");

            println("Private Shared Function _ruleNames_Extract() As String()");
            tabadd();
            println("Dim sb As New StringBuilder(256)");
            println("Dim s_1 As String");

            ids = grammar.rules.elements();
            int ruleNum = 0;
            int t_count1 = 0;
            tabadd();
            while (ids.hasMoreElements()) {
                GrammarSymbol sym = (GrammarSymbol) ids.nextElement();
                if (sym instanceof RuleSymbol) {
                    // println("  \"" + ( (RuleSymbol) sym).getId() + "\",");
                    println("s_1 = \"" + ( (RuleSymbol) sym).getId() + "\"");
                    if (t_count1 > 0 ) {println("sb.Append(\",\")");}
                    println("sb.Append(s_1)");
                }
                t_count1 ++;
            }
            // // println("};");
            println("Return sb.toString().split(\",\")");
             tabsub();
            // // println("}");
            println("End Function");
            println("");
        }

        // Generate nextToken() rule.
        // nextToken() is a synthetic lexer rule that is the implicit OR of all
        // user-defined lexer rules.
        genNextToken();

        // Generate code for each rule in the lexer
        ids = grammar.rules.elements();
        int ruleNum = 0;
        // Lexer Generating rules
        while (ids.hasMoreElements()) {
            RuleSymbol sym = (RuleSymbol) ids.nextElement();
            // Don't generate the synthetic rules
            if (!sym.getId().equals("mnextToken")) {
                genRule(sym, false, ruleNum++, grammar.tokenManager);
            }
            exitIfError();
        }

        // Generate the semantic predicate map for debugging
        if (grammar.debuggingOutput)
            genSemPredMap();

            // Generate the bitsets used throughout the lexer
        genBitsets(bitsetsUsed, ( (LexerGrammar) grammar).charVocabulary.size());

        println("");
         tabsub();
        // // println("}"); // end of a class -- Lexer?
        println("End Class");
         tabsub();
        // Generate the vb namespace closures (if required)
        if (nameSpace != null)
            nameSpace.emitClosures(currentOutput);

            // Close the lexer output stream
        currentOutput.close();
        currentOutput = null;
    }

    public void genInitFactory(Grammar g) {
        if (g.buildAST) {
            // Generate the method to initialize an ASTFactory when we're
            // building AST's
            // // println("static public void initializeASTFactory( ASTFactory factory )");
            println("Public Shared Sub initializeASTFactory( factory As ASTFactory  )");
            // // println("{");

            tabadd();
            // //println("factory.setMaxNodeType("+g.tokenManager.maxTokenType()+");");
            println("factory.setMaxNodeType(" + g.tokenManager.maxTokenType() +")");

            // Walk the token vocabulary and generate code to register every TokenID->ASTNodeType
            // mapping specified in the  tokens {...} section with the ASTFactory.
            Vector v = g.tokenManager.getVocabulary();
            for (int i = 0; i < v.size(); i++) {
                String s = (String) v.elementAt(i);
                if (s != null) {
                    TokenSymbol ts = g.tokenManager.getTokenSymbol(s);
                    if (ts != null && ts.getASTNodeType() != null) {
                        // // println("factory.setTokenTypeASTNodeType(" + s + ", \"" + ts.getASTNodeType() + "\");");
                        println("factory.setTokenTypeASTNodeType(" + s + ", \"" + ts.getASTNodeType() + "\")");
                    }
                }
            }

            tabsub();
            // // println("}");
            println("End Sub");
        }
    }

    public void genBody(ParserGrammar g) throws IOException {
        // Open the output stream for the parser and set the currentOutput
        // SAS: moved file setup so subclass could do it (for VAJ interface)
        setupOutput(grammar.getClassName());

        genAST = grammar.buildAST;

        // // tabs = 0;
        tabclear();
        // Generate the header common to all output files.
        genHeader();
        // Do not use printAction because we assume tabs==0
        println(behavior.getHeaderAction(""));

        // Generate the vb namespace declaration (if specified)
        /* For Vb, namesapce declaration is after Imports statements
        if (nameSpace != null)
            nameSpace.emitDeclarations(currentOutput);
        tabadd();
        */
        // Generate header for the parser
        // // println("// Generate the header common to all output files.");
        println("' ' // Generate the header common to all output files.");
        // // println("using System;");
        println("Imports System");
        println("Imports System.Text");
        println("");
        // // println("using TokenBuffer              = antlr.TokenBuffer;");
        // // // // println("Imports  TokenBuffer              = antlr.TokenBuffer");
        // // println("using TokenStreamException     = antlr.TokenStreamException;");
        // // // // println("Imports  TokenStreamException     = antlr.TokenStreamException");
        // // println("using TokenStreamIOException   = antlr.TokenStreamIOException;");
        // // // // println("Imports  TokenStreamIOException   = antlr.TokenStreamIOException");
        // // println("using ANTLRException           = antlr.ANTLRException;");
        // // // // println("Imports  ANTLRException           = antlr.ANTLRException");
        // // println("using " + grammar.getSuperClass() + " = antlr." + grammar.getSuperClass() + ";");
        // // // // println("Imports  " + grammar.getSuperClass() + " = antlr." + grammar.getSuperClass() );
        // // println("using Token                    = antlr.Token;");
        // // // // println("Imports  Token                    = antlr.Token");
        // // println("using TokenStream              = antlr.TokenStream;");
        // // // // println("Imports  TokenStream              = antlr.TokenStream");
        // // println("using RecognitionException     = antlr.RecognitionException;");
        // // // // println("Imports  RecognitionException     = antlr.RecognitionException");
        // // println("using NoViableAltException     = antlr.NoViableAltException;");
        // // // // println("Imports  NoViableAltException     = antlr.NoViableAltException");
        // // println("using MismatchedTokenException = antlr.MismatchedTokenException;");
        // // // // println("Imports  MismatchedTokenException = antlr.MismatchedTokenException");
        // // println("using SemanticException        = antlr.SemanticException;");
        // // // // println("Imports  SemanticException        = antlr.SemanticException");
        // // println("using ParserSharedInputState   = antlr.ParserSharedInputState;");
        // // // // println("Imports  ParserSharedInputState   = antlr.ParserSharedInputState");
        // // println("using BitSet                   = antlr.collections.impl.BitSet;");
        // // // // println("Imports  BitSet                   = antlr.collections.impl.BitSet");
        if (genAST) {
            // // println("using AST                      = antlr.collections.AST;");
            // // // // println("Imports AST                      = antlr.collections.AST");
            // // println("using ASTPair                  = antlr.ASTPair;");
            // // // // println("Imports ASTPair                  = antlr.ASTPair");
            // // println("using ASTFactory               = antlr.ASTFactory;");
            // // // // println("Imports ASTFactory               = antlr.ASTFactory");
            // // println("using ASTArray                 = antlr.collections.impl.ASTArray;");
            // // // // println("Imports ASTArray                 = antlr.collections.impl.ASTArray");
        }
        println("Imports  antlr");
        println("Imports  antlr.collections");
        println("Imports  antlr.collections.impl");
        println("Imports  antlr.debug");
        println("Imports  antlr.debug.misc");

        // Output the user-defined parser preamble
        println(grammar.preambleAction.getText());

        if (nameSpace != null){
                  nameSpace.emitDeclarations(currentOutput);
              }
        tabadd();
        println("");




        // Generate parser class definition
        String sup = null;
        if (grammar.superClass != null)
            sup = grammar.superClass;
        else
            sup = "antlr." + grammar.getSuperClass();

            // print javadoc comment if any
        if (grammar.comment != null) {
            _println(grammar.comment);
        }

        Token tprefix = (Token) grammar.options.get("classHeaderPrefix");
        if (tprefix == null) {
            print("Public ");
        }
        else {
            String p = StringUtils.stripFrontBack(tprefix.getText(), "\"", "\"");
            if (p == null) {
                print("Public ");
            }
            else {
                print(p + " ");
            }
        }

        // // println("class " + grammar.getClassName() + " : " + sup);
        println("Class " + grammar.getClassName());
        // //  + " : " + sup);
        println("Inherits " + sup);
        Token tsuffix = (Token) grammar.options.get("classHeaderSuffix");
        if (tsuffix != null) {
            String suffix = StringUtils.stripFrontBack(tsuffix.getText(), "\"",
                "\"");
            if (suffix != null)
                // // print("              , " + suffix);
                println("Implements " + suffix); //// must be an interface name for VB.Net
        }
        // // println("{");
        //\\ \\ just beginning of the parser class


        tabadd();

        // Generate 'const' definitions for Token IDs
        genTokenDefinitions(grammar.tokenManager);

        // set up an array of all the rule names so the debugger can
        // keep track of them only by number -- less to store in tree...
        if (grammar.debuggingOutput) {
            // println("private const string[] _ruleNames = {");
            println("Private Const _ruleNames() as String = {");

            tabadd();

            Enumeration ids = grammar.rules.elements();
            int ruleNum = 0;
            while (ids.hasMoreElements()) {
                GrammarSymbol sym = (GrammarSymbol) ids.nextElement();
                if (sym instanceof RuleSymbol)
                    // // println("  \"" + ( (RuleSymbol) sym).getId() + "\",");
                    println("  \"" + ( (RuleSymbol) sym).getId() + "\",");
            }
             tabsub();
            // // println("};");
            println("}");
        }

        // Generate user-defined parser class members
        print(
            processActionForSpecialSymbols(grammar.classMemberAction.getText(),
            grammar.classMemberAction.getLine(),
            currentRule, null)
            );

        // Generate parser class constructor from TokenBuffer
        println("");
        // // println("protected void initialize()");
        println("Protected Sub initialize()");
        // // println("{");

        tabadd();
        // // println("tokenNames = tokenNames_;");
        println("tokenNames = tokenNames_");

        if (grammar.buildAST) {
            // // println("initializeFactory();");
            println("initializeFactory()");
        }
            // if debugging, set up arrays and call the user-overridable
            //   debugging setup method
        if (grammar.debuggingOutput) {
            // // println("ruleNames  = _ruleNames;");
            println("ruleNames  = _ruleNames");
            // // println("semPredNames = _semPredNames;");
            println("semPredNames = _semPredNames");
            // // println("setupDebugging(tokenBuf);");
            println("setupDebugging(tokenBuf)");
        }
         tabsub();
        // // println("}");
        println("End Sub"); // end of initialize()
        println("");

        println("");
        // // println("protected " + grammar.getClassName() +"(TokenBuffer tokenBuf, int k) : base(tokenBuf, k)");
        // constructor
        println("Protected Sub New(tokenBuf As TokenBuffer , k As Integer)");
        // //  : base(tokenBuf, k)");

        tabadd();
        println("MyBase.New(tokenBuf, k)");
        // // println("{");
        // //   // did this by the aboved inserted
        // // println("initialize();");
        println("initialize()");
         tabsub();
        // // println("}");
        println("End Sub"); // end of the protected constructor
        println("");

        // Another (public ) constructor
        // // println("public " + grammar.getClassName() +"(TokenBuffer tokenBuf) : this(tokenBuf," + grammar.maxk + ")");
        println("Public Sub New(tokenBuf As TokenBuffer )");
        // // : this(tokenBuf," + grammar.maxk + ")");

        tabadd();
        println("MyClass.New(tokenBuf, " + grammar.maxk + ")");
        // // println("{");
         tabsub();
        // // println("}");
        println("End Sub");
        println("");

        // Generate parser class constructor from TokenStream
        // // println("protected " + grammar.getClassName() +"(TokenStream lexer, int k) : base(lexer,k)");
        println("Protected Sub New(lexer As TokenStream, k As Integer)");
        // // : base(lexer,k)");

        tabadd();
        println("MyBase.New(lexer, k)");
        // // println("{");
        // //
        // // println("initialize();");
        println("initialize()");
         tabsub();
        // // println("}");
        println("End Sub"); // end of the third constructor
        println("");

        // begin the fourth constructor
        // // println("public " + grammar.getClassName() + "(TokenStream lexer) : this(lexer," + grammar.maxk + ")");
        println("Public Sub New(lexer As TokenStream)");
        // // : this(lexer," + grammar.maxk + ")");

        tabadd();
        println("MyClass.New(lexer, " + grammar.maxk + ")");
        // // println("{");
        // // println("}");
         tabsub();
        // // println("}");
        println("End Sub");
        println("");

        // begin fifth constructor
        // // println("public " + grammar.getClassName() + "(ParserSharedInputState state) : base(state," + grammar.maxk +")");
        println("Public Sub New(state As ParserSharedInputState)");
        // //: base(state," + grammar.maxk +")");

        tabadd();
        println("MyBase.New(state, " + grammar.maxk + ")");
        // // println("{");
        // //
        // // println("initialize();");
        println("initialize()");
         tabsub();
        // // println("}");
        println("End Sub"); // end of fifth constructor
        println("");

        astTypes = new java.util.Vector(100);

        // Generate code for each rule in the grammar
        Enumeration ids = grammar.rules.elements();
        int ruleNum = 0;
        while (ids.hasMoreElements()) {
            GrammarSymbol sym = (GrammarSymbol) ids.nextElement();
            if (sym instanceof RuleSymbol) {
                RuleSymbol rs = (RuleSymbol) sym;
                genRule(rs, rs.references.size() == 0, ruleNum++,
                    grammar.tokenManager);
            }
            exitIfError();
        }
        if (usingCustomAST) {
            // when we are using a custom AST, overload Parser.getAST() to return the
            // custom AST type
            // // println("public new " + labeledElementASTType + " getAST()"); // ????  what the 'new' means here?
            println("Public Function getAST() As " + labeledElementASTType );
            // // println("{");

            // // println("return (" + labeledElementASTType + ") returnAST;");
            println("return DirectCast(returnAST, " + labeledElementASTType + ")");
             tabsub();
            // // println("}");
            println("End Function");
            println("");
        }

        // Generate the method that initializes the ASTFactory when we're
        // building AST's
        // // println("private void initializeFactory()");
        println("Private Sub initializeFactory()");
        // // println("{");

        tabadd();
        if (grammar.buildAST) {
            // // println("if (astFactory == null)");
            println("If (astFactory Is Nothing) Then");
            // // println("{");
             tabadd();
            if (usingCustomAST) {
                // // println("astFactory = new ASTFactory(\"" + labeledElementASTType + "\");");
                println("astFactory = new ASTFactory(\"" + labeledElementASTType + "\")");
            }
            else {
                // // println("astFactory = new ASTFactory();");
                println("astFactory = new ASTFactory()");
            }
             tabsub();
            // // println("}");
            println("End If"); // end of If
            // // println("initializeASTFactory( astFactory );");
            println("initializeASTFactory( astFactory )");
        }
         tabsub();
        // // println("}");
        println("End Sub"); // end of Sub initializeFactory
        genInitFactory(g);

        // Generate the token names
        genTokenStrings();

        // Generate the bitsets used throughout the grammar
        genBitsets(bitsetsUsed, grammar.tokenManager.maxTokenType());

        // Generate the semantic predicate map for debugging
        if (grammar.debuggingOutput)
            genSemPredMap();

            // Close class definition
        println("");
         tabsub();
        // // println("}");
        println("End Class"); // End of Class Parser

         tabsub();
        // Generate the vb namespace closures (if required)
        if (nameSpace != null)
            nameSpace.emitClosures(currentOutput);

            // Close the parser output stream
        currentOutput.close();
        currentOutput = null;
    }

    /*
     * About TreeWalker
     */
    public void genBody(TreeWalkerGrammar g) throws IOException {
        // Open the output stream for the parser and set the currentOutput
        // SAS: move file open to method so subclass can override it
        //      (mainly for VAJ interface)
        setupOutput(grammar.getClassName());

        genAST = grammar.buildAST;
        // // tabs = 0;
        tabclear();

        // Generate the header common to all output files.
        genHeader();
        // Do not use printAction because we assume tabs==0
        println(behavior.getHeaderAction(""));

        // Generate the vb namespace declaration (if specified)
        /* For VB, namespace declaration is after Imports statements
        if (nameSpace != null)
            nameSpace.emitDeclarations(currentOutput);
        tabadd();
        */

        // Generate header specific to the tree-parser VB file
        // // println("// Generate header specific to the tree-parser vb file");
        println("' ' // Generate header specific to the tree-parser vb file");

        // // println("using System;");
        // // println("");
        // // println("using " + grammar.getSuperClass() + " = antlr." + grammar.getSuperClass() + ";");
        // // println("using Token                    = antlr.Token;");
        // // println("using AST                      = antlr.collections.AST;");
        // // println("using RecognitionException     = antlr.RecognitionException;");
        // // println("using ANTLRException           = antlr.ANTLRException;");
        // // println("using NoViableAltException     = antlr.NoViableAltException;");
        // // println("using MismatchedTokenException = antlr.MismatchedTokenException;");
        // // println("using SemanticException        = antlr.SemanticException;");
        // // println("using BitSet                   = antlr.collections.impl.BitSet;");
        // // println("using ASTPair                  = antlr.ASTPair;");
        // // println("using ASTFactory               = antlr.ASTFactory;");
        // // println("using ASTArray                 = antlr.collections.impl.ASTArray;");

        println("Imports System");
        println("Imports System.Text");
        println("");
        // // // // println("Imports " + grammar.getSuperClass() + " = antlr." + grammar.getSuperClass() );
        // // // // println("Imports Token                    = antlr.Token");
        // // // // println("Imports AST                      = antlr.collections.AST");
        // // // // println("Imports RecognitionException     = antlr.RecognitionException");
        // // // // println("Imports ANTLRException           = antlr.ANTLRException");
        // // // // println("Imports NoViableAltException     = antlr.NoViableAltException");
        // // // // println("Imports MismatchedTokenException = antlr.MismatchedTokenException");
        // // // // println("Imports SemanticException        = antlr.SemanticException");
        // // // // println("Imports BitSet                   = antlr.collections.impl.BitSet");
        // // // // println("Imports ASTPair                  = antlr.ASTPair");
        // // // // println("Imports ASTFactory               = antlr.ASTFactory");
        // // // // println("Imports ASTArray                 = antlr.collections.impl.ASTArray");
        println("Imports  antlr");
        println("Imports  antlr.collections");
        println("Imports  antlr.collections.impl");
        println("Imports  antlr.debug");
        println("Imports  antlr.debug.misc");

        // Output the user-defined parser premamble
        println(grammar.preambleAction.getText());


        if (nameSpace != null){
            nameSpace.emitDeclarations(currentOutput);
        }
        tabadd();
        println("");


        // Generate parser class definition
        String sup = null;
        if (grammar.superClass != null) {
            sup = grammar.superClass;
        }
        else {
            sup = "antlr." + grammar.getSuperClass();
        }
        println("");

        // print javadoc comment if any
        if (grammar.comment != null) {
            _println(grammar.comment);
        }

        Token tprefix = (Token) grammar.options.get("classHeaderPrefix");
        if (tprefix == null) {
            // // print("public ");
            print("Public ");
        }
        else {
            String p = StringUtils.stripFrontBack(tprefix.getText(), "\"", "\"");
            if (p == null) {
                // // print("public ");
                print("Public ");
            }
            else {
                print(p + " ");
            }
        }

        // // println("class " + grammar.getClassName() + " : " + sup);
        println("Class " + grammar.getClassName());
        // //  + " : " + sup);

        tabadd();

        tabadd();

        tabadd();
        println("Inherits " + sup);

        Token tsuffix = (Token) grammar.options.get("classHeaderSuffix");
        if (tsuffix != null) {
            String suffix = StringUtils.stripFrontBack(tsuffix.getText(), "\"",
                "\"");
            if (suffix != null) {
                // // print("              , " + suffix);
                print("Implements " + suffix); // must be an interface name for VB
            }
        }
        println("");
         tabsub();
         tabsub();
         tabsub();
        // // println("{");

        tabadd();

        // Generate 'const' definitions for Token IDs
        genTokenDefinitions(grammar.tokenManager);

        // Generate user-defined parser class members
        print(
            processActionForSpecialSymbols(grammar.classMemberAction.getText(),
            grammar.classMemberAction.getLine(),
            currentRule, null)
            );

        // Generate default parser class constructor
        // // println("public " + grammar.getClassName() + "()");
        println("Public Sub New()");
        // // println("{");

        tabadd();
        // // println("tokenNames = tokenNames_;");
        println("tokenNames = tokenNames_");
         tabsub();
        // // println("}");
        println("End Sub"); // end of first constructor of the TreeWalker
        println("");

        astTypes = new java.util.Vector();
        // Generate code for each rule in the grammar
        Enumeration ids = grammar.rules.elements();
        int ruleNum = 0;
        String ruleNameInits = "";
        // Tree Walker, generate the bodies
        while (ids.hasMoreElements()) {
            GrammarSymbol sym = (GrammarSymbol) ids.nextElement();
            if (sym instanceof RuleSymbol) {
                RuleSymbol rs = (RuleSymbol) sym;

                genRule(rs, rs.references.size() == 0, ruleNum++,grammar.tokenManager);
            }
            exitIfError();
        }

        if (usingCustomAST) {
            // when we are using a custom ast override Parser.getAST to return the
            // custom AST type
            // // println("public new " + labeledElementASTType + " getAST()"); // what does the 'new' mean here??????
            println("Public Function getAST() As " + labeledElementASTType ); // what does the 'new' mean here??????
            // // println("{");

            tabadd();
            // // println("return (" + labeledElementASTType + ") returnAST;");
            println("return DirectCast(returnAST, " + labeledElementASTType + ")");
             tabsub();
            // // println("}");
            println("End Function");
            println("");
        }

        // Generate the ASTFactory initialization function
        genInitFactory(grammar);

        // Generate the token names
        genTokenStrings();

        // Generate the bitsets used throughout the grammar
        genBitsets(bitsetsUsed, grammar.tokenManager.maxTokenType());

        // Close class definition
         tabsub();
        // // println("}");
        println("End Class ' ' ' /// end of class TreeWalker"); // end of class TreeWalker
        println("");

         tabsub();
        // Generate the vb namespace closures (if required)
        if (nameSpace != null)
            nameSpace.emitClosures(currentOutput);

            // Close the parser output stream
        currentOutput.close();
        currentOutput = null;
    }

    /** Generate a series of case statements that implement a BitSet test.
     * @param p The Bitset for which cases are to be generated
     */
    protected void genCases(BitSet p) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genCases(" + p + ")");
        int[] elems;

        elems = p.toArray();
        // Wrap cases four-per-line for lexer, one-per-line for parser
        // // int wrap = (grammar instanceof LexerGrammar) ? 4 : 1;
        int wrap = (grammar instanceof LexerGrammar) ? 10 : 5;

        int j = 1;
        boolean startOfLine = true;
        // // StringBuffer sb = new StringBuffer(256);
        // // sb.append("Case ");
        if (elems.length < 1) {
            return;
        }


        print("Case " + getValueString(elems[0]));
        for (int i = 1; i < elems.length; i++) {
            _print(", ");
            if (j ==  wrap ) {_println(" _"); print("\t"); j=1;}else{j++;}
            _print(getValueString(elems[i]));
        }
        _println("");
        /*
        for (int i = 0; i < elems.length; i++) {
            if (j == 1) {
                print("");
                _print("case " + getValueString(elems[i]));
            }
            else {
                // // _print("  ");
                // // _print(", ");
                // // sb.append(", ");
                _print(", " + getValueString(elems[i]));
            }
            // // _print("case " + getValueString(elems[i]) + ":");
            // // _print("case " + getValueString(elems[i]));
            // // sb.append(getValueString(elems[i]));
            // // // // if (j == wrap) { _println(""); startOfLine = true; j = 1;}
            // // // // else {j++;startOfLine = false;}
        }
        // // _print(sb.toString());
        if (!startOfLine) {
            _println("");
        }
        */
    }

    /**Generate common code for a block of alternatives; return a
     * postscript that needs to be generated at the end of the
     * block.  Other routines may append else-clauses and such for
     * error checking before the postfix is generated.  If the
     * grammar is a lexer, then generate alternatives in an order
     * where alternatives requiring deeper lookahead are generated
     * first, and EOF in the lookahead set reduces the depth of
     * the lookahead.  @param blk The block to generate @param
     * noTestForSingle If true, then it does not generate a test
     * for a single alternative.
     */
    public VBBlockFinishingInfo genCommonBlock(AlternativeBlock blk,
        boolean noTestForSingle) {
        int nIF = 0;
        boolean createdLL1Switch = false;
        int closingBracesOfIFSequence = 0;
        VBBlockFinishingInfo finishingInfo = new VBBlockFinishingInfo();
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genCommonBlock(" + blk + ")");

            // Save the AST generation state, and set it to that of the block
        boolean savegenAST = genAST;
        genAST = genAST && blk.getAutoGen();

        boolean oldsaveTest = saveText;
        saveText = saveText && blk.getAutoGen();

        // Is this block inverted?  If so, generate special-case code
        if (blk.not &&
            analyzer.subruleCanBeInverted(blk, grammar instanceof LexerGrammar)) {
            if (DEBUG_CODE_GENERATOR)
                System.out.println("special case: ~(subrule)");
            Lookahead p = analyzer.look(1, blk);
            // Variable assignment for labeled elements
            if (blk.getLabel() != null && syntacticPredLevel == 0) {
                println(blk.getLabel() + " = " + lt1Value + ";");
            }

            // AST
            genElementAST(blk);

            String astArgs = "";
            if (grammar instanceof TreeWalkerGrammar) {
                if (usingCustomAST){
                    // // astArgs = "(AST)_t,";
                    astArgs = "DirectCast(_t, AST),";
                }
                else
                    astArgs = "_t,";
            }

            // match the bitset for the alternative
            // // println("match(" + astArgs + getBitsetName(markBitsetForGen(p.fset)) + ");");
            println("match(" + astArgs + getBitsetName(markBitsetForGen(p.fset)) + ")");

            // tack on tree cursor motion if doing a tree walker
            if (grammar instanceof TreeWalkerGrammar) {
                // // println("_t = _t.getNextSibling();");
                println("_t = _t.getNextSibling()");
            }
            return finishingInfo;
        }

        // Special handling for single alt
        if (blk.getAlternatives().size() == 1) {
            Alternative alt = blk.getAlternativeAt(0);
            // Generate a warning if there is a synPred for single alt.
            if (alt.synPred != null) {
                antlrTool.warning(
                    "Syntactic predicate superfluous for single alternative",
                    grammar.getFilename(),
                    blk.getAlternativeAt(0).synPred.getLine(),
                    blk.getAlternativeAt(0).synPred.getColumn()
                    );
            }
            if (noTestForSingle) {
                if (alt.semPred != null) {
                    // Generate validating predicate
                    genSemPred(alt.semPred, blk.line);
                }
                genAlt(alt, blk);
                return finishingInfo;
            }
        }

        // count number of simple LL(1) cases; only do switch for
        // many LL(1) cases (no preds, no end of token refs)
        // We don't care about exit paths for (...)*, (...)+
        // because we don't explicitly have a test for them
        // as an alt in the loop.
        //
        // Also, we now count how many unicode lookahead sets
        // there are--they must be moved to DEFAULT or ELSE
        // clause.
        int nLL1 = 0;
        for (int i = 0; i < blk.getAlternatives().size(); i++) {
            Alternative a = blk.getAlternativeAt(i);
            if (suitableForCaseExpression(a)) {
                nLL1++;
            }
        }

        // do LL(1) cases
        if (nLL1 >= makeSwitchThreshold) {
            // Determine the name of the item to be compared
            String testExpr = lookaheadString(1);
            createdLL1Switch = true;
            // when parsing trees, convert null to valid tree node with NULL lookahead
            if (grammar instanceof TreeWalkerGrammar) {
                // // println("if (null == _t)");
                println("If (_t is Nothing) Then ");

                tabadd();
                // // println("_t = ASTNULL;");
                println("_t = ASTNULL");
                tabsub();
                println("End If");
            }
            // // println("switch ( " + testExpr + " )");
            println("Select Case ( " + testExpr + " )"); // Selece
            // // println("{");

            tabadd();
            for (int i = 0; i < blk.alternatives.size(); i++) {

                Alternative alt = blk.getAlternativeAt(i);
                // ignore any non-LL(1) alts, predicated alts,
                // or end-of-token alts for case expressions
                bSaveIndexCreated = false;
                if (!suitableForCaseExpression(alt)) {
                    continue;
                }
                Lookahead p = alt.cache[1];
                if (p.fset.degree() == 0 && !p.containsEpsilon()) {
                    antlrTool.warning(
                        "Alternate omitted due to empty prediction set",
                        grammar.getFilename(),
                        alt.head.getLine(), alt.head.getColumn());
                }
                else {
                    genCases(p.fset);
                    // // println("{");

                    tabadd();
                    genAlt(alt, blk);
                    // // println("break;"); // No need 'break' for VB
                    tabsub();
                    // // println("}");
                    // // println("End Select"); --- // !?! end of this case block, not the End Select
                }
            }
            // // println("default:");
            println("Case Else");

            tabadd();
        }

        // do non-LL(1) and nondeterministic cases This is tricky in
        // the lexer, because of cases like: STAR : '*' ; ASSIGN_STAR
        // : "*="; Since nextToken is generated without a loop, then
        // the STAR will have end-of-token as it's lookahead set for
        // LA(2).  So, we must generate the alternatives containing
        // trailing end-of-token in their lookahead sets *after* the
        // alternatives without end-of-token.  This implements the
        // usual lexer convention that longer matches come before
        // shorter ones, e.g.  "*=" matches ASSIGN_STAR not STAR
        //
        // For non-lexer grammars, this does not sort the alternates
        // by depth Note that alts whose lookahead is purely
        // end-of-token at k=1 end up as default or else clauses.
        int startDepth = (grammar instanceof LexerGrammar) ? grammar.maxk : 0;
        boolean isHasIfInner = false; boolean isNeedEndIf =false;
        for (int altDepth = startDepth; altDepth >= 0; altDepth--) {
            if (DEBUG_CODE_GENERATOR)
                System.out.println("checking depth " + altDepth);
            for (int i = 0; i < blk.alternatives.size(); i++) {
                Alternative alt = blk.getAlternativeAt(i);

                if (DEBUG_CODE_GENERATOR)
                    System.out.println("genAlt: " + i);
                    // if we made a switch above, ignore what we already took care
                    // of.  Specifically, LL(1) alts with no preds
                    // that do not have end-of-token in their prediction set
                    // and that are not giant unicode sets.
                if (createdLL1Switch && suitableForCaseExpression(alt)) {
                    if (DEBUG_CODE_GENERATOR)
                        System.out.println(
                            "ignoring alt because it was in the switch");
                    continue;
                }
                String e;

                boolean unpredicted = false;

                if (grammar instanceof LexerGrammar) {
                    // Calculate the "effective depth" of the alt,
                    // which is the max depth at which
                    // cache[depth]!=end-of-token
                    int effectiveDepth = alt.lookaheadDepth;
                    if (effectiveDepth == GrammarAnalyzer.NONDETERMINISTIC) {
                        // use maximum lookahead
                        effectiveDepth = grammar.maxk;
                    }
                    while (effectiveDepth >= 1 &&
                        alt.cache[effectiveDepth].containsEpsilon()) {
                        effectiveDepth--;
                    }
                    // Ignore alts whose effective depth is other than
                    // the ones we are generating for this iteration.
                    if (effectiveDepth != altDepth) {
                        if (DEBUG_CODE_GENERATOR)
                            System.out.println(
                                "ignoring alt because effectiveDepth!=altDepth;" +
                                effectiveDepth + "!=" + altDepth);
                        continue;
                    }
                    unpredicted = lookaheadIsEmpty(alt, effectiveDepth);
                    e = getLookaheadTestExpression(alt, effectiveDepth);


                }
                else {
                    unpredicted = lookaheadIsEmpty(alt, grammar.maxk);
                    e = getLookaheadTestExpression(alt, grammar.maxk);
                }

                // Was it a big unicode range that forced unsuitability
                // for a case expression?
                if (alt.cache[1].fset.degree() > caseSizeThreshold &&
                    suitableForCaseExpression(alt)) {
                    if (nIF == 0) {
                        // // println("if " + e);
                        println("if " + e + " Then");
                        // // println("{");
                    }
                    else {
                        // // println("else if " + e);
                        println("ElseIf " + e + " Then");
                        // // println("{");
                    }
                    isHasIfInner = true;
                }
                else if (unpredicted &&
                    alt.semPred == null &&
                    alt.synPred == null) {
                    // The alt has empty prediction set and no
                    // predicate to help out.  if we have not
                    // generated a previous if, just put {...} around
                    // the end-of-token clause
                    if (nIF == 0) {
                        // // println("{");
                    }
                    else {
                        // // println("else {");
                        println("Else ");
                        isNeedEndIf = true;
                    }
                    finishingInfo.needAnErrorClause = false;
                }
                else {


                    // check for sem and syn preds
                    // Add any semantic predicate expression to the lookahead test
                    if (alt.semPred != null) {
                        // if debugging, wrap the evaluation of the predicate in a method
                        //
                        // translate $ and # references
                        ActionTransInfo tInfo = new ActionTransInfo();
                        String actionStr = processActionForSpecialSymbols(alt.
                            semPred,
                            blk.line,
                            currentRule,
                            tInfo);
                        // ignore translation info...we don't need to
                        // do anything with it.  call that will inform
                        // SemanticPredicateListeners of the result
                        if ( ( (grammar instanceof ParserGrammar) ||
                            (grammar instanceof LexerGrammar)) &&
                            grammar.debuggingOutput) {
                            // // e = "(" + e + "&& fireSemanticPredicateEvaluated(antlr.debug.SemanticPredicateEvent.PREDICTING," + //FIXME
                            // //     addSemPred(charFormatter.escapeString(actionStr)) +
                            // //     "," + actionStr + "))";
                            e = "(" + e + "AndAlso fireSemanticPredicateEvaluated(antlr.debug.SemanticPredicateEvent.PREDICTING," + //FIXME
                                addSemPred(charFormatter.escapeString(actionStr)) +
                                "," + actionStr + "))";


                        }
                        else {

                            e = "(" + e + " AndAlso (" + actionStr + "))";
                        }
                    }

                    // Generate any syntactic predicates
                    if (nIF > 0) {
                        if (alt.synPred != null) {
                            // // println("else {");
                            println("Else ");
                            isNeedEndIf = true;
                            tabadd();
                            genSynPred(alt.synPred, e);
                            closingBracesOfIFSequence++;
                        }
                        else {
                            // // println("else if " + e + " {");
                            println("ElseIf " + e + " Then");
                        }
                    }
                    else {
                        if (alt.synPred != null) {
                            genSynPred(alt.synPred, e);
                        }
                        else {
                            // when parsing trees, convert null to valid tree node
                            // with NULL lookahead.
                            if (grammar instanceof TreeWalkerGrammar) {
                                // // println("if (_t == null)");
                                println("If (_t Is Nothing) Then");

                                tabadd();
                                // // println("_t = ASTNULL;");
                                println("_t = ASTNULL");
                                 tabsub();
                                println("End If");
                            }
                            // // println("if " + e);
                            println("If " + e + " Then");
                            isHasIfInner = true;
                            // // println("{");
                        }
                    }

                }

                nIF++;

                tabadd();
                genAlt(alt, blk);
                 tabsub();
                // // println("}");
                // // println("End If");
                if (isHasIfInner && isNeedEndIf) {
                    println("End If");
                    isHasIfInner  = false;
                    isNeedEndIf = false;
                }

            }

        }


        String ps = "";
        // // // // for (int i = 1; i <= closingBracesOfIFSequence; i++) {
        // // // //     // // ps += "}";
        // // // //     ps += "End If\n"; // // !?!
        // // // // }
        // // if (nIF > 0 ) {
        // //       println("End If"); // // This is added
        // // }
        // Restore the AST generation state
        genAST = savegenAST;

        // restore save text state
        saveText = oldsaveTest;

        // Return the finishing info.
        if (createdLL1Switch) {
             tabsub(); tabsub();
            // // finishingInfo.postscript = ps + "break; }";
            finishingInfo.postscript = ps + "End Select "; // // !?! x?x
            finishingInfo.generatedSwitch = true;
            finishingInfo.generatedAnIf = nIF > 0;
            //return new VBBlockFinishingInfo(ps+"}",true,nIF>0); // close up switch statement
        }
        else {
            finishingInfo.postscript = ps;
            finishingInfo.generatedSwitch = false;
            finishingInfo.generatedAnIf = nIF > 0;
            // return new VBBlockFinishingInfo(ps, false,nIF>0);
        }
        return finishingInfo;
    }

    private static boolean suitableForCaseExpression(Alternative a) {
        return a.lookaheadDepth == 1 &&
            a.semPred == null &&
            !a.cache[1].containsEpsilon() &&
            a.cache[1].fset.degree() <= caseSizeThreshold;
    }

    /** Generate code to link an element reference into the AST */
    private void genElementAST(AlternativeElement el) {
        // handle case where you're not building trees, but are in tree walker.
        // Just need to get labels set up.
        if (grammar instanceof TreeWalkerGrammar && !grammar.buildAST) {
            String elementRef;
            String astName;

            // Generate names and declarations of the AST variable(s)
            if (el.getLabel() == null) {
                elementRef = lt1Value;
                // Generate AST variables for unlabeled stuff
                astName = "tmp" + astVarNumber + "_AST";
                astVarNumber++;
                // Map the generated AST variable in the alternate
                mapTreeVariable(el, astName);
                // Generate an "input" AST variable also
                // // println(labeledElementASTType + " " + astName + "_in = " + elementRef + ";");
                println("Dim " + astName + "_in As " + labeledElementASTType + " = " + elementRef );
            }
            return;
        }

        if (grammar.buildAST && syntacticPredLevel == 0) {
            boolean needASTDecl =
                (genAST &&
                (el.getLabel() != null ||
                (el.getAutoGenType() != GrammarElement.AUTO_GEN_BANG)));

            // RK: if we have a grammar element always generate the decl
            // since some guy can access it from an action and we can't
            // peek ahead (well not without making a mess).
            // I'd prefer taking this out.
            if (el.getAutoGenType() != GrammarElement.AUTO_GEN_BANG &&
                (el instanceof TokenRefElement))
                needASTDecl = true;

            boolean doNoGuessTest = (grammar.hasSyntacticPredicate &&
                needASTDecl);

            String elementRef;
            String astNameBase;

            // Generate names and declarations of the AST variable(s)
            if (el.getLabel() != null) {
                // if the element is labeled use that name...
                elementRef = el.getLabel();
                astNameBase = el.getLabel();
            }
            else {
                // else generate a temporary name...
                elementRef = lt1Value;
                // Generate AST variables for unlabeled stuff
                astNameBase = "tmp" + astVarNumber;
                astVarNumber++;
            }

            // Generate the declaration if required.
            if (needASTDecl) {
                // Generate the declaration
                if (el instanceof GrammarAtom) {
                    GrammarAtom ga = (GrammarAtom) el;
                    if (ga.getASTNodeType() != null) {
                        genASTDeclaration(el, astNameBase, ga.getASTNodeType());
                        //println(ga.getASTNodeType()+" " + astName+" = null;");
                    }
                    else {
                        genASTDeclaration(el, astNameBase,
                            labeledElementASTType);
                        //println(labeledElementASTType+" " + astName + " = null;");
                    }
                }
                else {
                    genASTDeclaration(el, astNameBase, labeledElementASTType);
                    //println(labeledElementASTType+" " + astName + " = null;");
                }
            }

            // for convenience..
            String astName = astNameBase + "_AST";

            // Map the generated AST variable in the alternate
            mapTreeVariable(el, astName);
            if (grammar instanceof TreeWalkerGrammar) {
                // Generate an "input" AST variable also
                // // println(labeledElementASTType + " " + astName + "_in = null;");
                println("Dim " + astName + "_in As labeledElementASTType = null");
            }

            // Enclose actions with !guessing
            if (doNoGuessTest) {
                //println("if (0 == inputState.guessing)");
                //println("{");
                //
            }

            // if something has a label assume it will be used
            // so we must initialize the RefAST
            if (el.getLabel() != null) {
                if (el instanceof GrammarAtom) {
                    // // println(astName + " = " + getASTCreateString( (GrammarAtom) el, elementRef) +";");
                    println(astName + " = " + getASTCreateString( (GrammarAtom) el, elementRef) );
                }
                else {
                    // // println(astName + " = " + getASTCreateString(elementRef) +";");
                    println(astName + " = " + getASTCreateString(elementRef) );
                }
            }

            // if it has no label but a declaration exists initialize it.
            if (el.getLabel() == null && needASTDecl) {
                elementRef = lt1Value;
                if (el instanceof GrammarAtom) {
                    // // println(astName + " = " +getASTCreateString( (GrammarAtom) el, elementRef) +";");
                    println(astName + " = " +getASTCreateString( (GrammarAtom) el, elementRef) );
                }
                else {
                    // // println(astName + " = " + getASTCreateString(elementRef) + ";");
                    println(astName + " = " + getASTCreateString(elementRef) );
                }
                // Map the generated AST variable in the alternate
                if (grammar instanceof TreeWalkerGrammar) {
                    // set "input" AST variable also
                    // // println(astName + "_in = " + elementRef + ";");
                    println(astName + "_in = " + elementRef );
                }
            }

            if (genAST) {
                switch (el.getAutoGenType()) {
                    case GrammarElement.AUTO_GEN_NONE:
                        if (usingCustomAST ||
                            ( (el instanceof GrammarAtom) &&
                            ( ( (GrammarAtom) el).getASTNodeType() != null))){
                            // // println("astFactory.addASTChild(currentAST, (AST)" + astName + ");");
                            println("astFactory.addASTChild(currentAST, DirectCast(" + astName +", AST)");
                        }
                        else{
                            // // println("astFactory.addASTChild(currentAST, " + astName + ");");
                            println("astFactory.addASTChild(currentAST, " + astName + ")");
                        }
                        break;
                    case GrammarElement.AUTO_GEN_CARET:
                        if (usingCustomAST ||
                            ( (el instanceof GrammarAtom) &&
                            ( ( (GrammarAtom) el).getASTNodeType() != null))) {
                            // // println("astFactory.makeASTRoot(currentAST, (AST)" + astName + ");");
                            println("astFactory.makeASTRoot(currentAST, DirectCast(" + astName +", AST)");
                        }
                        else{
                            // // println("astFactory.makeASTRoot(currentAST, " + astName + ");");
                            println("astFactory.makeASTRoot(currentAST, " + astName + ")");
                        }
                        break;
                    default:
                        break;
                }
            }
            if (doNoGuessTest) {
                // tabsub();
                //println("}");
            }
        }
    }

    /** Close the try block and generate catch phrases
     * if the element has a labeled handler in the rule
     */
    private void genErrorCatchForElement(AlternativeElement el) {
        if (el.getLabel() == null)
            return;
        String r = el.enclosingRuleName;
        if (grammar instanceof LexerGrammar) {
            r = CodeGenerator.encodeLexerRuleName(el.enclosingRuleName);
        }
        RuleSymbol rs = (RuleSymbol) grammar.getSymbol(r);
        if (rs == null) {
            antlrTool.panic("Enclosing rule not found!");
        }
        ExceptionSpec ex = rs.block.findExceptionSpec(el.getLabel());
        if (ex != null) {
             tabsub();
            // // println("}"); // !?! should disable the "}" before catch clause?
            genErrorHandler(ex);
        }
    }

    /** Generate the catch phrases for a user-specified error handler */
    private void genErrorHandler(ExceptionSpec ex) {
        // Each ExceptionHandler in the ExceptionSpec is a separate catch
        for (int i = 0; i < ex.handlers.size(); i++) {
            ExceptionHandler handler = (ExceptionHandler) ex.handlers.elementAt(
                i);
            // Generate catch phrase
            // // println("catch (" + handler.exceptionTypeAndName.getText() + ")");
            println("Catch   " + handler.exceptionTypeAndName.getText() );
            // // println("{");

            tabadd();
            if (grammar.hasSyntacticPredicate) {
                // // println("if (0 == inputState.guessing)");
                println("If ( inputState.guessing = 0 ) Then ");
                // // println("{");

                tabadd();
            }

            // When not guessing, execute user handler action
            ActionTransInfo tInfo = new ActionTransInfo();
            printAction(processActionForSpecialSymbols(handler.action.getText(),
                handler.action.getLine(), currentRule, tInfo));

            if (grammar.hasSyntacticPredicate) {
                 tabsub();
                // // println("}");
                // // println("else");
                println("Else");
                // // println("{");

                tabadd();
                // When guessing, rethrow exception
                // // println("throw " + extractIdOfAction(handler.exceptionTypeAndName) + ";");
                // // println("throw;");
                String exName = handler.exceptionTypeAndName.getText();
                int pos_ = exName.lastIndexOf(" ");
                if (pos_>0 ) {
                    exName = exName.substring(pos_+1);
                }
                // // // // println("Throw " + exName);
                println("Throw " + extractIdOfAction(handler.exceptionTypeAndName));
                 tabsub();
                // // println("}");
                println("End If");
            }
            // Close catch phrase
             tabsub();
            // // println("}");
            println("End Try");
        }
    }

    /** Generate a try { opening if the element has a labeled handler in the rule */
    private void genErrorTryForElement(AlternativeElement el) {
        if (el.getLabel() == null)
            return;
        String r = el.enclosingRuleName;
        if (grammar instanceof LexerGrammar) {
            r = CodeGenerator.encodeLexerRuleName(el.enclosingRuleName);
        }
        RuleSymbol rs = (RuleSymbol) grammar.getSymbol(r);
        if (rs == null) {
            antlrTool.panic("Enclosing rule not found!");
        }
        ExceptionSpec ex = rs.block.findExceptionSpec(el.getLabel());
        if (ex != null) {
            // // println("try   // for error handling");
            println("Try   ' '  // for error handling");
            // // println("{");

            tabadd();
        }
    }

    protected void genASTDeclaration(AlternativeElement el) {
        genASTDeclaration(el, labeledElementASTType);
    }

    protected void genASTDeclaration(AlternativeElement el, String node_type) {
        genASTDeclaration(el, el.getLabel(), node_type);
    }

    protected void genASTDeclaration(AlternativeElement el, String var_name,
        String node_type) {
        // already declared?
        if (declaredASTVariables.contains(el))
            return;

        // emit code
        //String s = StringUtils.stripFrontBack(node_type, "\"", "\"");
        //println(s + " " + var_name + "_AST = null;");
        // // println(node_type + " " + var_name + "_AST = null;");
        println("Dim " + var_name + "_AST As " + node_type + " = Nothing");

        // mark as declared
        declaredASTVariables.put(el, el);
    }

    /** Generate a header that is common to all vb files */
    protected void genHeader() {
        // // println("// $ANTLR " + Tool.version + ": " +
        // //     "\"" + antlrTool.fileMinusPath(antlrTool.grammarFile) + "\"" +
        // //     " -> " +
        // //     "\"" + grammar.getClassName() + ".cs\"$");
        println("' ' ' ********************************************");
        println("' ' ' $ANTLR " + Tool.version + ": " +
            "\"" + antlrTool.fileMinusPath(antlrTool.grammarFile) + "\"" +
            " -> " +
            "\"" + grammar.getClassName() + ".vb\"$");
        println("' ' ' ********************************************");
    }

    private void genLiteralsTest() {
        // // println("_ttype = testLiteralsTable(_ttype);");
        println("_ttype = testLiteralsTable(_ttype)");
    }

    private void genLiteralsTestForPartialToken() {
        // // println("_ttype = testLiteralsTable(text.ToString(_begin, text.Length-_begin), _ttype);");
        println("_ttype = testLiteralsTable(text.ToString(_begin, text.Length-_begin), _ttype)");
    }

    protected void genMatch(BitSet b) {
    }

    protected void genMatch(GrammarAtom atom) {
        if (atom instanceof StringLiteralElement) {
            if (grammar instanceof LexerGrammar) {
                genMatchUsingAtomText(atom);
            }
            else {
                genMatchUsingAtomTokenType(atom);
            }
        }
        else if (atom instanceof CharLiteralElement) {
            if (grammar instanceof LexerGrammar) {
                genMatchUsingAtomText(atom);
            }
            else {
                antlrTool.error("cannot ref character literals in grammar: " +
                    atom);
            }
        }
        else if (atom instanceof TokenRefElement) {
            genMatchUsingAtomText(atom);
        }
        else if (atom instanceof WildcardElement) {
            gen( (WildcardElement) atom);
        }
    }
    /*
    protected String convertToVBstring(String ins) {
        if (ins == null || ins.length()<1) {return  "";}
        StringBuffer sb = new StringBuffer(256);
        MessageFormat mf = new MessageFormat("{0}");
        String tt = "d\"n";
        Object[] res=null;
        try {
             res = mf.parse(tt);
        }catch (Exception e) {
            tt = e.getMessage();
        }
        return sb.toString();
    }
    */

   // Only process Char literal string with front and end single quotes
   protected String convertCharLiteralPresentationToVB(String ins) {
       if (ins.indexOf("\\u")>-1) {
           int stp2=03;
       }
       if (ins == null || ins.length() <2 ) {return ins;}
       if (ins.equals("\"\\r\\n\"")) return "vbCrLf";
       if (ins.charAt(0) !='\'' || ins.charAt(ins.length()-1) !='\'' ) { return ins;}
       String s = ins.substring(1, ins.length()-1);
       if (ins.equals("\'\"\'") ) {
           int stpHere = 111;
       }
       if (s.charAt(0) != '\\') {
           switch(s.charAt(0)) {
               case '\"': return "\"\"\"\"c";
               default: return "\"" + s.substring(0) + "\"c";
           }
       }
       switch (s.charAt(1)){
           case 'n': return "vbLf";
           case 'r': return "vbCr";
           case 't':
               return "vbTab";
           case '\"': return "\"\"\"\"c";
           case '\\': return "\"\\\"c";
           case 'u':
               return convertUnicode2VBCharString(s.substring(2));
           case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7':
               return convertOct2VBCharString(s.substring(1));
           case 'x': return convertHex2DecimalString(s.substring(2)); // should only be two Hex digits
           default:
               return "\""+s.substring(1)+"\"c";
       }
   }

    protected String convertToVBstring(String ins) {
        if (ins == null || ins.length()<1) {return  "";}
        StringBuffer sb = new StringBuffer(256);
        boolean isEsc, isOct, isHex, isUnic, isCtrl, isCharQuoted, isStrQuoted, isEscString;
        boolean isCrntDigit;
        int index, count, i, j, k;
        char ch, ch2, ch3; String quote = ""; String quoteBreakL=""; String quoteBreakR="";
        isEsc = false; isOct = false; isHex = false; isUnic = false; isCtrl = false; isCharQuoted = false; isStrQuoted = false; isEscString = false;
        index = 0; i = 0; j = 0; k = 0;
        count = ins.length();
        index = 0;
        if (count >1 && (ins.charAt(0) == '\'' && ins.charAt(count-1) == '\'')) {
            count = count - 2 ; index = 1; isCharQuoted = true; quote ="\"";
        }else if (count >1 && (ins.charAt(0) == '"' && ins.charAt(count-1) == '"')) {
            count = count - 2 ; index = 1; isStrQuoted = true; quote = "\"";
        }
        quoteBreakL = "\" & "; quoteBreakR = " & \"";
        StringBuffer s = new StringBuffer(16);
        String t_s = "";
        isCrntDigit = false;
        for (; index < count-1; index++) {
            ch = ins.charAt(index);
            switch (ch) {
                case '\\':
                    if (isEsc) {
                        s.append( "\\"); isEsc = false; isEscString = true;
                        if (isCharQuoted) {return quote + s.toString() + quote + "c";}
                    }
                    else { isEsc = true; }
                    break;
                case 'n':
                    if (isEsc) {
                        s.append("vbLf"); isEsc = false; isEscString = true;
                        if (isCharQuoted) {return s.toString() ;}
                    }
                    else { isEsc = true; }
                    break;
                case 'r':
                    if (isEsc) {
                        s.append( "vbCr"); isEsc = false; isEscString = true;
                        if (isCharQuoted) {return  s.toString() ;}
                    }
                    else { isEsc = true; }
                    break;
                case 't':
                    if (isEsc) {
                        s.append( "vbTab"); isEsc = false; isEscString = true;
                        if (isCharQuoted) {return  s.toString() ;}
                    }
                    else { isEsc = true; }
                    break;
                case '0': case '1': case '2': case '3': case '4': case '5':case '6':case '7':
                    isCrntDigit = true;
                    if (isEsc) { isOct = true; isEsc = false;}
                    if (isOct) {
                        s.append(ch);
                        if (s.length()>2) {
                            t_s = this.convertOct2VBCharString(s.toString());
                            s.delete(0,s.length());
                            s.append(t_s);
                            isEscString = true; isOct = false;
                        }
                        if (isCharQuoted) {return  s.toString() ;}
                    }else if (isHex) { // Hex decimal byte
                        s.append(ch);
                        if (s.length()>1) {
                            t_s = getVBEscCharString(this.convertHex2DecimalString(s.toString()));
                            s.delete(0,s.length());
                            s.append(t_s);
                            isEscString = true; isHex =false;
                        }
                        if (isCharQuoted) {return  s.toString() ;}
                    }else if(isUnic){
                        s.append(ch);
                       if (s.length()>1) {
                           t_s = getVBEscCharString(this.convertHex2DecimalString(s.toString()));
                           s.delete(0,s.length());
                           s.append(t_s);
                           isEscString = true; isUnic =false;
                       }
                       if (isCharQuoted) {return  s.toString() ;}
                    }else {
                        s.append(ch);
                    }
                    break;
                case '8':case '9':
                    isCrntDigit = true;
                    if (isHex) { // Hex decimal byte
                        s.append(ch);
                        if (s.length()>1) {
                            t_s = getVBEscCharString(this.convertHex2DecimalString(s.toString()));
                            s.delete(0, s.length());
                            s.append(t_s);
                            isEscString = true; isHex =false;
                        }
                        if (isCharQuoted) {return  s.toString() ;}
                    }else if(isUnic){
                        s.append(ch);
                        if (s.length() > 3) {
                            t_s = getVBEscCharString(this.convertHex2DecimalString(s.toString()));
                            s.delete(0, s.length());
                            s.append(t_s);
                            isEscString = true;
                            isUnic = false;
                        }
                        if (isCharQuoted) {
                            return s.toString();
                        }
                    }
                    else {
                        s.append(ch);
                    }
                break;

                case 'a':case 'A': case 'b': case 'B': case 'c': case 'C': case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
                    if (isHex) {isCrntDigit = true;}
                case 'x':
                case 'u':

                default:

            } // end of switch

            goodExit:
                return sb.toString();
        } // end of outter for

        return sb.toString();
    }

    protected String convertOct2VBCharString(String oct) {
        if (oct == null || oct.length() !=3) {return oct;}
        int pos;
        int charIntVal = Integer.parseInt(oct.substring(0,1))*64 + Integer.parseInt(oct.substring(1,1))*8+Integer.parseInt(oct.substring(2,1));
        return "Chr(" + charIntVal + ")";
    }
    protected String convertUnicode2VBCharString(String unic) { // unic is a four digits hex decimal number string
        if (unic == null || unic.length() !=4) {return unic;}
        unic = convertHex2DecimalString(unic);
        return "Chr(" + unic + ")";
    }
    protected String getVBEscCharString(String EscNumber) {
        if (EscNumber == null || EscNumber.length() !=4) {return EscNumber;}
        return "Char(" + EscNumber + ")";
    }
    protected String convertHex2DecimalString(String hex) {
        if (hex == null || hex.length() <1) {return hex;}
        int res=0; int unit = 16;
        char ch ;
        for (int pos = 0; pos< hex.length(); pos++){
            ch = hex.charAt(0);
            res = res * unit;
            switch (ch) {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    res  += Integer.parseInt("" + ch);
                    break;
                case 'a':
                case 'A':
                    res += 10;
                    break;
                case 'b':
                case 'B':
                    res += 11;
                    break;
                case 'c':
                case 'C':
                    res += 12;
                    break;
                case 'd':
                case 'D':
                    res += 13;
                    break;
                case 'e':
                case 'E':
                    res += 14;
                    break;
                case 'f':
                case 'F':
                    res += 15;
                    break;
                default:
                    res += 0;
            } // switch
        } // for
    return ""+res;
}


    protected void genMatchUsingAtomText(GrammarAtom atom) {
        // match() for trees needs the _t cursor
        String astArgs = "";
        if (grammar instanceof TreeWalkerGrammar) {
            if (usingCustomAST)
                astArgs = "(AST)_t,";
            else
                astArgs = "_t,";
        }

        // if in lexer and ! on element, save buffer index to kill later
        if (grammar instanceof LexerGrammar &&
            (!saveText || atom.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
            declareSaveIndexVariableIfNeeded();
            // // println("_saveIndex = text.Length;");
            println("_saveIndex = text.Length");
        }


        print(atom.not ? "matchNot(" : "match(");

        _print(convertCharLiteralPresentationToVB(astArgs));
         // // / / / / /this.convertToVBstring(astArgs);
        // print out what to match
        if (atom.atomText.equals("EOF")) {
            // horrible hack to handle EOF case
            _print("Token.EOF_TYPE");
        }
        else {
            // // _print(atom.atomText);
            //if (atom.atomText.indexOf("\\r\\n" )>-1 ) {
            //    int stphere2=2;
            //    stphere2++;
            //}

            _print(convertCharLiteralPresentationToVB(atom.atomText));
        }
        // // _println(");");
        _println(")");

        if (grammar instanceof LexerGrammar &&
            (!saveText || atom.getAutoGenType() == GrammarElement.AUTO_GEN_BANG)) {
            declareSaveIndexVariableIfNeeded();
            // // println("text.Length = _saveIndex;"); // kill text atom put in buffer
            println("text.Length = _saveIndex");
        }
    }

    protected void genMatchUsingAtomTokenType(GrammarAtom atom) {
        // match() for trees needs the _t cursor
        String astArgs = "";
        if (grammar instanceof TreeWalkerGrammar) {
            if (usingCustomAST)
                astArgs = "(AST)_t,";
            else
                astArgs = "_t,";
        }

        // If the literal can be mangled, generate the symbolic constant instead
        String mangledName = null;
        String s = astArgs + getValueString(atom.getType());

        // matching
        // // println( (atom.not ? "matchNot(" : "match(") + s + ");");
        println( (atom.not ? "matchNot(" : "match(") + s + ")");
    }

    /** Generate the nextToken() rule.  nextToken() is a synthetic
     * lexer rule that is the implicit OR of all user-defined
     * lexer rules.
     */
    public void genNextToken() {
        // Are there any public rules?  If not, then just generate a
        // fake nextToken().
        boolean hasPublicRules = false;
        for (int i = 0; i < grammar.rules.size(); i++) {
            RuleSymbol rs = (RuleSymbol) grammar.rules.elementAt(i);
            if (rs.isDefined() && rs.access.equals("public")) {
                hasPublicRules = true;
                break;
            }
        }
        if (!hasPublicRules) {
            println("");
            // // println("override public Token nextToken()\t\t\t//throws TokenStreamException");
            println("Public Overrides Function nextToken() As Token ' ' ' \t\t\t//throws TokenStreamException");
            // // println("{");
             tabadd();
            // // println("try");
            println("Try");
            // // println("{");
             tabadd();
            // // println("uponEOF();");
            println("uponEOF()");
             tabsub();
            // // println("}");
            // // println("catch(CharStreamIOException csioe)");
            println("Catch  csioe As CharStreamIOException ");
            // // println("{");
             tabadd();
            // // println("throw new TokenStreamIOException(csioe.io);");
            println("throw new TokenStreamIOException(csioe.io)");
             tabsub();
            // // println("}");
            // // println("catch(CharStreamException cse)");
            println("Catch  cse As CharStreamException ");
            // // println("{");
             tabadd();
            // // println("throw new TokenStreamException(cse.Message);");
            println("throw new TokenStreamException(cse.Message)");
             tabsub();
            // // println("}");
            println("End Try");
            // // println("return new CommonToken(Token.EOF_TYPE, \"\");");
            println("return new CommonToken(Token.EOF_TYPE, \"\")");
             tabsub();
            // // println("}");
            println("End Function");
            println("");
            return;
        }


        /*
         *
         */
        // Create the synthesized nextToken() rule
        RuleBlock nextTokenBlk = MakeGrammar.createNextTokenRule(grammar,
            grammar.rules, "nextToken");
        // Define the nextToken rule symbol
        RuleSymbol nextTokenRs = new RuleSymbol("mnextToken");
        nextTokenRs.setDefined();
        nextTokenRs.setBlock(nextTokenBlk);
        nextTokenRs.access = "private";
        grammar.define(nextTokenRs);
        // Analyze the nextToken rule
        boolean ok = grammar.theLLkAnalyzer.deterministic(nextTokenBlk);

        // Generate the next token rule
        String filterRule = null;
        if ( ( (LexerGrammar) grammar).filterMode) {
            filterRule = ( (LexerGrammar) grammar).filterRule;
        }

        println("");
        // // println("override public Token nextToken()\t\t\t//throws TokenStreamException");
        println("Public Overrides Function nextToken() As Token ' ' ' \t\t\t//throws TokenStreamException");
        // // println("{");
         tabadd();
        // // println("Token theRetToken = null;");
        println("Dim theRetToken As Token ");
        _println("TryAgain:");
        // // println("for (;;)");
        println("While (True)");
        // // println("{");
        tabadd();
        // // println("Token _token = null;");
        println("Dim _token As Token ");
        // // println("int _ttype = Token.INVALID_TYPE;");
        println("Dim _ttype As Integer = Token.INVALID_TYPE");
        if ( ( (LexerGrammar) grammar).filterMode) {
            // // println("setCommitToPath(false);");
            println("setCommitToPath(false)");
            if (filterRule != null) {
                // Here's a good place to ensure that the filter rule actually exists
                if (!grammar.isDefined(CodeGenerator.encodeLexerRuleName(
                    filterRule))) {
                    grammar.antlrTool.error("Filter rule " + filterRule + " does not exist in this lexer");
                }
                else {
                    RuleSymbol rs = (RuleSymbol) grammar.getSymbol(
                        CodeGenerator.encodeLexerRuleName(filterRule));
                    if (!rs.isDefined()) {
                        grammar.antlrTool.error("Filter rule " + filterRule +
                            " does not exist in this lexer");
                    }
                    else if (rs.access.equals("public")) {
                        grammar.antlrTool.error("Filter rule " + filterRule +
                            " must be protected");
                    }
                }
                // // println("int _m;");
                println("Dim _m As Integer");
                // // println("_m = mark();");
                println("_m = mark()");
            }
        }
        // // println("resetText();");
        println("resetText()");

        // // println("try     // for char stream error handling");
        println("Try     ' ' ' // for char stream error handling");
        // // println("{");
         tabadd();

        // Generate try around whole thing to trap scanner errors
        // // println("try     // for lexical error handling");
        println("Try     ' ' ' // for lexical error handling");
        // // println("{");
        tabadd();

        // Test for public lexical rules with empty paths
        for (int i = 0; i < nextTokenBlk.getAlternatives().size(); i++) {
            Alternative a = nextTokenBlk.getAlternativeAt(i);
            if (a.cache[1].containsEpsilon()) {
                //String r = a.head.toString();
                RuleRefElement rr = (RuleRefElement) a.head;
                String r = CodeGenerator.decodeLexerRuleName(rr.targetRule);
                antlrTool.warning("public lexical rule " + r +
                    " is optional (can match \"nothing\")");
            }
        }

        // Generate the block // in the generating next token

        VBBlockFinishingInfo howToFinish = genCommonBlock(nextTokenBlk, false);
        tabadd();
        // // String errFinish ="if (LA(1)==EOF_CHAR) { uponEOF(); returnToken_ = makeToken(Token.EOF_TYPE); }";
        String errFinish ="If (LA(1) = EOF_CHAR) Then" +newline +
            tabstr+"\t" + "uponEOF()" +newline +
            tabstr+"\t" + "returnToken_ = makeToken(Token.EOF_TYPE)";
        errFinish += newline + tabstr;
        if ( ( (LexerGrammar) grammar).filterMode) {

            if (filterRule == null) {
                //kunle: errFinish += "else { consume(); continue tryAgain; }";
                // // errFinish += "\t\t\t\telse";
                errFinish += "\t\t\t\tElse";
                // // errFinish += "\t\t\t\t{";
                // // errFinish += "\t\t\t\t\tconsume();";
                errFinish += "\t\t\t\t\tconsume()";
                // // errFinish += "\t\t\t\t\tgoto tryAgain;";
                errFinish += "\t\t\t\t\tgoto tryAgain";
                // // errFinish += "\t\t\t\t}";
                errFinish += "\t\t\t\tEnd If";
            }
            else {
                // // errFinish += "\t\t\t\t\telse" + newline +
                // //     "\t\t\t\t\t{" + newline +
                // //     "\t\t\t\t\tcommit();" + newline +
                // //     "\t\t\t\t\ttry {m" + filterRule + "(false);}" + newline +
                // //     "\t\t\t\t\tcatch(RecognitionException e)" + newline +
                // //     "\t\t\t\t\t{" + newline +
                // //     "\t\t\t\t\t	// catastrophic failure" + newline +
                // //     "\t\t\t\t\t	reportError(e);" + newline +
                // //     "\t\t\t\t\t	consume();" + newline +
                // //     "\t\t\t\t\t}" + newline +
                // //     "\t\t\t\t\tgoto tryAgain;" + newline +
                 // //    "\t\t\t\t}";
                errFinish += "\t\t\t\t\tElse" + newline +
                    // // "\t\t\t\t\t{" + newline +
                    "\t\t\t\t\tcommit()" + newline +
                    "\t\t\t\t\tTry "+ newline+ "\t\t\t\t\t\tm" + filterRule + "(false)" + newline +
                    "\t\t\t\t\tCatch  e  As RecognitionException  " + newline +
                    // // "\t\t\t\t\t{" + newline +
                    "\t\t\t\t\t	' ' // catastrophic failure" + newline +
                    "\t\t\t\t\t	reportError(e)" + newline +
                    "\t\t\t\t\t	consume()" + newline +
                    "\t\t\t\t\tEnd Try" + newline +
                    "\t\t\t\t\tgoto tryAgain" + newline +
                    "\t\t\t\tEnd If";
            }
        }
        else {
            // // errFinish += "else {" + throwNoViable + "}";

            errFinish += "Else" + newline + tabstr+ "\t" + throwNoViable + newline + tabstr+  "End If";
        }
        tabsub();
        genBlockFinish(howToFinish, errFinish);

        // at this point a valid token has been matched, undo "mark" that was done
        if ( ( (LexerGrammar) grammar).filterMode && filterRule != null) {
            // // println("commit();");
            println("commit()");
        }

        // Generate literals test if desired
        // make sure _ttype is set first; note returnToken_ must be
        // non-null as the rule was required to create it.
        // // println("if ( null==returnToken_ ) goto tryAgain; // found SKIP token");
        println("If ( returnToken_ Is Nothing ) Then goto tryAgain ' ' // found SKIP token");
        // // println("_ttype = returnToken_.Type;");
        println("_ttype = returnToken_.Type");
        if ( ( (LexerGrammar) grammar).getTestLiterals()) {
            genLiteralsTest();
        }

        // return token created by rule reference in switch
        // // println("returnToken_.Type = _ttype;");
        println("returnToken_.Type = _ttype");
        // // println("return returnToken_;");
        println("return returnToken_");

        // Close try block
        tabsub();
        // // println("}");
        // // println("catch (RecognitionException e) {");
        println("Catch  e  As  RecognitionException  ");
        tabadd();
        if ( ( (LexerGrammar) grammar).filterMode) {
            if (filterRule == null) {
                // // println("if (!getCommitToPath())");
                println("If (Not getCommitToPath()) Then ");
                // // println("{");
                 tabadd();
                // // println("consume();");
                println("consume()");
                // // println("goto tryAgain;");
                println("GoTo tryAgain");
                 tabsub();
                // // println("}");
                println("End If");
            }
            else {
                // // println("if (!getCommitToPath())");
                println("If (Not getCommitToPath()) Then");
                // // println("{");
                 tabadd();
                // // println("rewind(_m);");
                println("rewind(_m)");
                // // println("resetText();");
                println("resetText()");
                // // println("try {m" + filterRule + "(false);}");
                println("Try");
                // // {m" + filterRule + "(false);}");
                 tabadd();
                println("m" + filterRule + "(false)");
                 tabsub();
                // // println("catch(RecognitionException ee) {");
                println("Catch ee As RecognitionException");
                // // println("	// horrendous failure: error in filter rule");
                tabadd();
                println("' ' // horrendous failure: error in filter rule");
                // // println("	reportError(ee);");
                println("reportError(ee)");
                // // println("consume();");
                println("consume()");
                // // println("}");
                //println("goto tryAgain;");
                tabsub();
                // // println("}");
                println("End Try");
                tabsub();
                println("else");
            }
        }
        if (nextTokenBlk.getDefaultErrorHandler()) {
            // // println("{");
             tabadd();
             // // println("reportError(e);");
             println("reportError(e)");
             // // println("consume();");
             println("consume()");
             tabsub();
            // // println("}");
        }
        else {
            // pass on to invoking routine
            tabadd();
            // // println("throw new TokenStreamRecognitionException(e);");
            println("Throw new TokenStreamRecognitionException(e)");
            tabsub();
        }
        tabsub();
        // // println("}"); // end of Catch (RecognitionException e)

        // close CharStreamException try
        // // println("}");
        println("End Try"); // end of Try --- closure of Catch (RecognitionException e) -- Inner Try for // for lexical error handling
        tabsub();
        // // println("catch (CharStreamException cse) {");
        println("Catch  cse As CharStreamException ");
        tabadd();
        // // println(" if ( cse is CharStreamIOException ) {");
        println("If ( Typeof cse Is CharStreamIOException ) Then");
        tabadd();
        // // println("throw new TokenStreamIOException(((CharStreamIOException)cse).io);");
        println("Throw new TokenStreamIOException( DirectCast(cse, CharStreamIOException).io)");
        // // println("	}");
        tabsub();
        // // println(" else {");
        println("Else ");
        tabadd();
        // // println(" throw new TokenStreamException(cse.Message);");
        println("Throw new TokenStreamException(cse.Message)");
        tabsub();
        // // println("	}");
        println("End If");
        tabsub();
        println("End Try"); // End of all Try (outter Try)

        // close for-loop
        tabsub();
        // // println("}");
        println("End While");

        // close method nextToken
        tabsub();
        // // println("}");
        println("End Function");
        println("");
    }


    protected String strRepeat(String src, int repeatingNum) {
        StringBuffer sb=new StringBuffer(256);
        for (int i=0; i<repeatingNum; i++) {
            sb.append(src);
        }
        return sb.toString();
    }

    protected String checkTypeString(String shouldBeCTypeString) {
        int pos;
        pos = shouldBeCTypeString.indexOf("float");
        if (pos >= 0) {
            String s = shouldBeCTypeString.replaceAll("float", "Single");
            return s;
        }
        return shouldBeCTypeString;
    }

    protected String ensurefrontDIM(String funcReturnStatement) {
        int pos;
        if (funcReturnStatement == null || funcReturnStatement.length() < 1) {
            return funcReturnStatement;
        }
        funcReturnStatement = funcReturnStatement.trim();
        if (!funcReturnStatement.substring(0, 4).toLowerCase().equals("dim ")) {
            funcReturnStatement = "Dim " + funcReturnStatement;
        }
        return funcReturnStatement;
    }


/*
    protected String convtFuncFormalParamToVB(String formalParamSstring) {
        if (formalParamSstring == null || formalParamSstring.length() < 1) { return "";}
        StringBuffer sb=new StringBuffer(256);
        int i, i2, j, count; String[] t= formalParamSstring.split(",");
        if (t == null || t.length<1  ) {
            return formalParamSstring;
        }
        j=0;
        for (i = 0; i < t.length; i++) {
            t[i]=t[i].trim();
            if (t[i].length() > 0) {
                j++;
            }
            if (i > j) {
                t[j] = t[i];
            }
        }
        String[] t2; j=0;
        for (i=0; i< t.length; i++) {
            t2 = t[i].split(" ");
            for (i2 = 0; i2< t2.length; i2++) {
                t2[i2]=t2[i2].trim();
                if (t2[i2].length()>0 ) {j++;}
                if (i2>j) {t2[j]=t2[i2];}
                if (j>3) {
                     if (i>0) {sb.append(", ");}
                    sb.append(t2[1]);
                    sb.append(" As ");
                    sb.append(t2[3]);
                    break;
                }
            }
         }
         return sb.toString();
    }
*/

protected String[] splitString(String s, String delimiter) {
    String[] res, t_s;
    StringBuffer buf = new StringBuffer(256);
    int i, j, count;
    if (s == null) { return new String[0]; }
    t_s = s.split(delimiter);
    if ( t_s == null || t_s.length < 1 ) { return new String[]{s};}
    j = 0 ;
    count = t_s.length;
    for (i = 0; i < count; i++) {

    }
    return null;
}


/*
protected String convtFuncFormalParamToVB(String formalParamSstring) {
    if (formalParamSstring == null || formalParamSstring.length() < 1) { return "";}
    StringBuffer sb=new StringBuffer(256);
    int i, i2, j, count; String[] t= formalParamSstring.split(",");
    if (t == null || t.length<1  ) {
        return formalParamSstring;
    }
    j=0;
    for (i = 0; i < t.length; i++) {
        t[i]=t[i].trim();
        if (t[i].length() > 0) {
            j++;
        }
        if (i > j) {
            t[j] = t[i];
        }
    }
    String[] t2; j=0;
    for (i=0; i< t.length; i++) {
        t2 = t[i].split(" ");
        for (i2 = 0; i2< t2.length; i2++) {
            t2[i2]=t2[i2].trim();
            if (t2[i2].length()>0 ) {j++;}
            if (i2>j) {t2[j]=t2[i2];}
            if (j>3) {
                 if (i>0) {sb.append(", ");}
                sb.append(t2[1]);
                sb.append(" As ");
                sb.append(t2[3]);
                break;
            }
        }
     }
     return sb.toString();
}

 */
// Assume original syntax: ([Dim] env As SCPCommonEnv, [Dim] optional byref vNode As ExprNode = Nothing)
protected String convtFuncFormalParamToVB(String formalParamSstring) {
    if (formalParamSstring == null || formalParamSstring.length() < 1) { return "";}
    StringBuffer sb=new StringBuffer(256);
    int i, i2, j, count;
    String[] t;
    String[] t1= formalParamSstring.split(",");
    if (t1 == null || t1.length<1  ) {
        return formalParamSstring;
    }
    j=0;
    for (i = 0; i < t1.length; i++) {
        t1[i]=t1[i].trim();
        if (t1[i].length() > 0) {
            j++;
        }
        if (i > j) {
            t1[j] = t1[i];
        }
    }
    if (i == j)
        t = t1;
    else { // then j must less than i
        t = new String[j];
        for (i=0; i<t.length; i++) {
            t[i] = t1[i];
        }
    }



    String[] t2; j=0;
    int pos=-1;
    // processing every t1
    for (i=0; i<t.length; i++) {
        if (t[i].substring(0,4).toLowerCase().equals("dim ")) {
            t[i] = t[i].substring(4);
        }else {
            pos = t[i].toLowerCase().indexOf(" dim ");
            if ( pos> 0) {
                t[i] = t[i].substring(0, pos) + " " + t[i].substring(pos+5);
            }
        }
        if (i>0) sb.append(", ");
        sb.append(t[i]);
    }
    return sb.toString();
}


    /*
     * overrides parent -- CodeGenerator's extractIdOfAction method
     */
    protected String extractIdOfAction(String s, int line, int column) {
        String t;
        t = s.toLowerCase();
        int t_pos = t.lastIndexOf(" as ");
        int len;
        char ch;
        if (t_pos < 0) { //not found type info, just return original
            // Something is bogus, but we cannot parse the language-specific
            // actions any better.  The compiler will have to catch the problem.
            antlrTool.warning("Ill-formed action", grammar.getFilename(), line, column);
            return s;
        }
        t_pos -= 1;
        len = s.length();
        StringBuffer sb = new StringBuffer(128);
        while (t_pos >=0) {
            ch = s.charAt(t_pos);
            if (Character.isJavaIdentifierPart(ch)) {
                sb.append(ch);
            }
            else {
                break;
            }
            t_pos--;
        }
        sb.reverse();
        return sb.toString();
    }


    /*
     * overrides parent -- CodeGenerator's extractTypeOfAction method
     */
    protected String extractTypeOfAction(String s, int line, int column) {
        // s = removeAssignmentFromDeclaration(s);
        // Search back from the end for a non alphanumeric.  That marks the
        // beginning of the identifier
        String t;
        t = s.toLowerCase();
        int t_pos = t.lastIndexOf(" as ");
        int len; char ch;
        if (t_pos < 0 ) { //not found type info, just return original
            antlrTool.warning("Ill-formed action", grammar.getFilename(), line, column);
            return s;
        }
        t_pos += 4;
        len = s.length();
        StringBuffer sb = new StringBuffer(128);
        while (t_pos < len) {
            ch = s.charAt(t_pos);
            if (Character.isJavaIdentifierPart(ch)){
                sb.append(ch);
            }else {
                break;
            }
            t_pos++;
        }
        return sb.toString();
    }


    /*
    protected String changeReturnActionTextToVB(String shouldBeReturnActionString) {
        String[] t = shouldBeReturnActionString.split(" ");
        int i,j, count;
        if (t == null || t.length<1  ) {
            return shouldBeReturnActionString;
            //res = "Dim " + shouldBeReturnActionString.substring(pos+1) + " As " + shouldBeReturnActionString.subSequence(0,pos);
        }
        j=0;
        for (i = 0; i< t.length; i++) {
            if (i<3) {
                if (t[i].length() > 0) {j++;}
                if (i > j) {t[j] = t[i];}
            }else{
                t[2] += t[i];
            }
        }
        switch (j) {
            case 1: return shouldBeReturnActionString;
            case 2: return "Dim " + t[1] + " As " + t[0];
            case 3: return "Dim " + t[1] + " As " + t[0] + " " + t[2];
        }
        return shouldBeReturnActionString;
    }
    */


    /** Gen a named rule block.
     * ASTs are generated for each element of an alternative unless
     * the rule or the alternative have a '!' modifier.
     *
     * If an alternative defeats the default tree construction, it
     * must set <rule>_AST to the root of the returned AST.
     *
     * Each alternative that does automatic tree construction, builds
     * up root and child list pointers in an ASTPair structure.
     *
     * A rule finishes by setting the returnAST variable from the
     * ASTPair.
     *
     * @param rule The name of the rule to generate
     * @param startSymbol true if the rule is a start symbol (i.e., not referenced elsewhere)
     */
    public void genRule(RuleSymbol s, boolean startSymbol, int ruleNum,
        TokenManager tm) {
        tabs = 1;
        if (DEBUG_CODE_GENERATOR)
            System.out.println("genRule(" + s.getId() + ")");
        if (!s.isDefined()) {
            antlrTool.error("undefined rule: " + s.getId());
            return;
        }

        // Generate rule return type, name, arguments
        RuleBlock rblk = s.getBlock();
        currentRule = rblk;
        currentASTResult = s.getId();

        // clear list of declared ast variables..
        declaredASTVariables.clear();

        // Save the AST generation state, and set it to that of the rule
        boolean savegenAST = genAST;
        genAST = genAST && rblk.getAutoGen();

        // boolean oldsaveTest = saveText;
        saveText = rblk.getAutoGen();

        // print javadoc comment if any
        if (s.comment != null) {
            _println(s.comment);
        }

        // Gen method access and final qualifier
        //print(s.access + " final ");
        print(s.access + " ");

        // Gen method return type (note lexer return action set at rule creation)


        if (rblk.returnAction != null) {
            // Has specified return value
            // moved to later as per VB format
            // // _print(extractTypeOfAction(rblk.returnAction, rblk.getLine(), rblk.getColumn()) + " ");
            _print(" Function ");
        }
        else {
            // No specified return value
            // // _print("void ");
            _print(" Sub ");
        }


        // Gen method name
        _print(s.getId() + "(");

        // Additional rule parameters common to all rules for this grammar
        // // _print(commonExtraParams);
         _print(convtFuncFormalParamToVB(commonExtraParams));
        //_print(commonExtraParams);
        if (commonExtraParams.length() != 0 && rblk.argAction != null) {
            _print(",");
        }

        // Gen arguments
        if (rblk.argAction != null) {
            // Has specified arguments
            // // _println("");
             tabadd();
            // // println(rblk.argAction);
            // // _print(convtFuncFormalParamToVB(rblk.argAction));
            _print(rblk.argAction);
             tabsub();
            // // print(")");
            _print(")");
        }
        else {
            // No specified arguments
            _print(")");
        }

        /*
         * This is added part for adjusting to VB's format
         */
        if (rblk.returnAction != null) {
            // Function has a return value
            _print(" As "); // // added
            _print(this.checkTypeString( extractTypeOfAction(rblk.returnAction, rblk.getLine(), rblk.getColumn())) + " ");
        }
        else {
            // No specified return value
            // // _print("void ");
            // // If it is a Sub, then empty here
        }



        // Gen throws clause and open curly
        // // _print(" //throws " + exceptionThrown);
        _print("' ' //throws " + exceptionThrown);
        if (grammar instanceof ParserGrammar) {
            _print(", TokenStreamException");
        }
        else if (grammar instanceof LexerGrammar) {
            _print(", CharStreamException, TokenStreamException");
        }
        // Add user-defined exceptions unless lexer (for now)
        if (rblk.throwsSpec != null) {
            if (grammar instanceof LexerGrammar) {
                antlrTool.error(
                    "user-defined throws spec not allowed (yet) for lexer rule " +
                    rblk.ruleName);
            }
            else {
                _print(", " + rblk.throwsSpec);
            }
        }

        _println("");
        // // _println("{"); take off this block
         tabadd(); // This is the function's starting ident

        // Convert return action to variable declaration
        if (rblk.returnAction != null){
            // // println(rblk.returnAction + ";");
            // // println(this.checkTypeString(changeReturnActionTextToVB(rblk.returnAction)) );
            println(ensurefrontDIM(checkTypeString(rblk.returnAction)));

        }

            // print out definitions needed by rules for various grammar types
        println(commonLocalVars);

        if (grammar.traceRules) {
            if (grammar instanceof TreeWalkerGrammar) {
                if (usingCustomAST){
                    // // println("traceIn(\"" + s.getId() + "\",(AST)_t);");
                    println("traceIn(\"" + s.getId() + "\",(AST)_t)");
                }
                else{
                    // // println("traceIn(\"" + s.getId() + "\",_t);");
                    println("traceIn(\"" + s.getId() + "\",_t)");
                }
            }
            else {
                // // println("traceIn(\"" + s.getId() + "\");");
                println("traceIn(\"" + s.getId() + "\")");
            }
        }

        if (grammar instanceof LexerGrammar) {
            // lexer rule default return value is the rule's token name
            // This is a horrible hack to support the built-in EOF lexer rule.
            if (s.getId().equals("mEOF")){
                // // println("_ttype = Token.EOF_TYPE;");
                println("_ttype = Token.EOF_TYPE");
            }
            else{
                // // println("_ttype = " + s.getId().substring(1) + ";");
                println("_ttype = " + s.getId().substring(1) );
            }


                // delay creation of _saveIndex until we need it OK?
            bSaveIndexCreated = false;

            /*
                  println("boolean old_saveConsumedInput=saveConsumedInput;");
                 if ( !rblk.getAutoGen() ) {		// turn off "save input" if ! on rule
                  println("saveConsumedInput=false;");
                  }
             */
        }

        // if debugging, write code to mark entry to the rule
        if (grammar.debuggingOutput)
            if (grammar instanceof ParserGrammar){
                // // println("fireEnterRule(" + ruleNum + ",0);");
                println("fireEnterRule(" + ruleNum + ",0)");
            }
            else if (grammar instanceof LexerGrammar){
                // // println("fireEnterRule(" + ruleNum + ",_ttype);");
                println("fireEnterRule(" + ruleNum + ",_ttype)");
            }

                // Generate trace code if desired
        if (grammar.debuggingOutput || grammar.traceRules) {
            // // println("try { // debugging");
            println("Try ' '  // debugging");
            tabadd();
        }

        // Initialize AST variables
        if (grammar instanceof TreeWalkerGrammar) {
            // "Input" value for rule
            // // println(labeledElementASTType + " " + s.getId() + "_AST_in = (" + labeledElementASTType + ")_t;");
            println("Dim " + s.getId() + "_AST_in As " + labeledElementASTType + " = directCast(_t, " + labeledElementASTType + ")");
        }
        if (grammar.buildAST) {
            // Parser member used to pass AST returns from rule invocations
            // // println("returnAST = null;");
            println("returnAST = Nothing");
            // Tracks AST construction
            // println("ASTPair currentAST = (inputState.guessing==0) ? new ASTPair() : null;");
            // // println("ASTPair currentAST = new ASTPair();");
            println("Dim currentAST As ASTPair  = new ASTPair()");
            // User-settable return value for rule.
            // // println(labeledElementASTType + " " + s.getId() + "_AST = null;");
            println("Dim " + s.getId() + "_AST As " + labeledElementASTType + " = Nothing");
        }

        genBlockPreamble(rblk);
        genBlockInitAction(rblk);
        println("");

        // Search for an unlabeled exception specification attached to the rule
        ExceptionSpec unlabeledUserSpec = rblk.findExceptionSpec("");

        // Generate try block around the entire rule for  error handling
        if (unlabeledUserSpec != null || rblk.getDefaultErrorHandler()) {
            // // println("try {      // for error handling");
            println("Try  ' ' // for error handling");
            tabadd();
        }

        // Generate the alternatives
        if (rblk.alternatives.size() == 1) {
            // One alternative -- use simple form
            Alternative alt = rblk.getAlternativeAt(0);
            String pred = alt.semPred;
            if (pred != null)
                genSemPred(pred, currentRule.line);
            if (alt.synPred != null) {
                antlrTool.warning(
                    "Syntactic predicate ignored for single alternative",
                    grammar.getFilename(), alt.synPred.getLine(),
                    alt.synPred.getColumn()
                    );
            }
            // // // // if (s.getId().substring(1).equals("CHAR_LITERAL_sym") ) {
            // // // //     int lookAthere = 111; // for testing
            // // // // }
            if (s.getId().substring(1).equals("ESC") ) {
                int lookAthere = 111; // for testing
            }

            genAlt(alt, rblk);
        }
        else {
            // Multiple alternatives -- generate complex form
            boolean ok = grammar.theLLkAnalyzer.deterministic(rblk);

            VBBlockFinishingInfo howToFinish = genCommonBlock(rblk, false);
            genBlockFinish(howToFinish, throwNoViable);
        }

        // Generate catch phrase for error handling
        if (unlabeledUserSpec != null || rblk.getDefaultErrorHandler()) {
            // Close the try block
            tabsub();
            // // println("}"); // end of Try block (Not of Try Catch End Try block)
        }

        // Generate user-defined or default catch phrases
        if (unlabeledUserSpec != null) {
            genErrorHandler(unlabeledUserSpec);
        }
        else if (rblk.getDefaultErrorHandler()) {
            // Generate default catch phrase
            // /// println("catch (" + exceptionThrown + " ex)");
            println("Catch   ex As " + exceptionThrown );
            // // println("{");
            tabadd();
            // Generate code to handle error if not guessing
            if (grammar.hasSyntacticPredicate) {
                // // println("if (0 == inputState.guessing)");
                println("If ( inputState.guessing = 0 ) Then ");
                // // println("{");
                tabadd();
            }
            // // println("reportError(ex);");
            println("reportError(ex)");
            if (! (grammar instanceof TreeWalkerGrammar)) {
                // Generate code to consume until token in k==1 follow set
                Lookahead follow = grammar.theLLkAnalyzer.FOLLOW(1,
                    rblk.endNode);
                String followSetName = getBitsetName(markBitsetForGen(follow.
                    fset));
                // // println("consume();");
                println("consume()");
                // // println("consumeUntil(" + followSetName + ");");
                println("consumeUntil(" + followSetName + ")");
            }
            else {
                // Just consume one token
                // // println("if (null != _t)");
                println("If (Not _t Is Nothing ) Then");
                // // println("{");
                tabadd();
                // // println("_t = _t.getNextSibling();");
                println("_t = _t.getNextSibling()");
                tabsub();
                // // println("}");
                println("End If");
            }
            if (grammar.hasSyntacticPredicate) {
                 tabsub();
                // When guessing, rethrow exception
                // // println("}");
                // // println("else");
                println("Else");
                // // println("{");
                tabadd();
                //println("throw ex;");
                // // println("throw;");
                println("Throw ex");
                tabsub();
                // // println("}");
                println("End If");
            }
            // Close catch phrase
            tabsub();
            // // println("}");
            println("End Try");
        }

        // Squirrel away the AST "return" value
        if (grammar.buildAST) {
            // // println("returnAST = " + s.getId() + "_AST;");
            println("returnAST = " + s.getId() + "_AST");
        }

        // Set return tree value for tree walkers
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("retTree_ = _t;");
            println("retTree_ = _t");
        }

        // Generate literals test for lexer rules so marked
        if (rblk.getTestLiterals()) {
            if (s.access.equals("protected")) {
                genLiteralsTestForPartialToken();
            }
            else {
                genLiteralsTest();
            }
        }

        println(""); // // an empty line
        // if doing a lexer rule, dump code to create token if necessary
        if (grammar instanceof LexerGrammar) {
            // // println("if (_createToken && (null == _token) && (_ttype != Token.SKIP))");
            println("If (_createToken AndAlso ( _token Is Nothing ) AndAlso (_ttype <> Token.SKIP)) Then ");
            // // println("{");
            tabadd();
            // // println("_token = makeToken(_ttype);");
            println("_token = makeToken(_ttype)");
            // // println("_token.setText(text.ToString(_begin, text.Length-_begin));");
            println("_token.setText(text.ToString(_begin, text.Length-_begin))");
            tabsub();
            // // println("}");
            println("End If");
            // // println("returnToken_ = _token;");
            println("returnToken_ = _token");
        }

        // Gen the return statement if there is one (lexer has hard-wired return action)
        if (rblk.returnAction != null) {
            // // println("return " + extractIdOfAction(rblk.returnAction, rblk.getLine(), rblk.getColumn()) + ";");
            println("return " + extractIdOfAction(rblk.returnAction, rblk.getLine(), rblk.getColumn()) );
        }

        if (grammar.debuggingOutput || grammar.traceRules) {
            tabsub();
            // // println("}");
            // // println("finally");
            println("Finally");
            // // println("{ // debugging");
            tabadd();

            // If debugging, generate calls to mark exit of rule
            if (grammar.debuggingOutput)
                if (grammar instanceof ParserGrammar){
                    // // println("fireExitRule(" + ruleNum + ",0);");
                    println("fireExitRule(" + ruleNum + ",0)");
                }
                else if (grammar instanceof LexerGrammar){
                    // // println("fireExitRule(" + ruleNum + ",_ttype);");
                    println("fireExitRule(" + ruleNum + ",_ttype)");
                }

            if (grammar.traceRules) {
                if (grammar instanceof TreeWalkerGrammar) {
                    // // println("traceOut(\"" + s.getId() + "\",_t);");
                    println("traceOut(\"" + s.getId() + "\",_t)");
                }
                else {
                    // // println("traceOut(\"" + s.getId() + "\");");
                    println("traceOut(\"" + s.getId() + "\")");
                }
            }

            tabsub();
            // // println("}");
            println("End Try"); // End of outter Try -- debugging
        }

        tabsub();
        // // println("}");
        // End of this function definition
        if (rblk.returnAction != null) {
            // Has specified return value
            // moved to later as per VB format
            // // _print(extractTypeOfAction(rblk.returnAction, rblk.getLine(), rblk.getColumn()) + " ");
            print("End Function");
        }
        else {
            // No specified return value
            // // _print("void ");
            print("End Sub");
        }
        println("");
        println("");
        println("");

        // Restore the AST generation state
        genAST = savegenAST;

        // restore char save state
        // saveText = oldsaveTest;
    }

    private void GenRuleInvocation(RuleRefElement rr) {
        // dump rule name
        _print(rr.targetRule + "(");

        // lexers must tell rule if it should set returnToken_
        if (grammar instanceof LexerGrammar) {
            // if labeled, could access Token, so tell rule to create
            if (rr.getLabel() != null) {
                _print("True");
            }
            else {
                _print("false");
            }
            if (commonExtraArgs.length() != 0 || rr.args != null) {
                _print(",");
            }
        }

        // Extra arguments common to all rules for this grammar
        _print(commonExtraArgs);
        if (commonExtraArgs.length() != 0 && rr.args != null) {
            _print(",");
        }

        // Process arguments to method, if any
        RuleSymbol rs = (RuleSymbol) grammar.getSymbol(rr.targetRule);
        if (rr.args != null) {
            // When not guessing, execute user arg action
            ActionTransInfo tInfo = new ActionTransInfo();
            String args = processActionForSpecialSymbols(rr.args, 0,
                currentRule, tInfo);
            if (tInfo.assignToRoot || tInfo.refRuleRoot != null) {
                antlrTool.error("Arguments of rule reference '" + rr.targetRule +
                    "' cannot set or ref #" +
                    currentRule.getRuleName(), grammar.getFilename(),
                    rr.getLine(), rr.getColumn());
            }
            _print(args);

            // Warn if the rule accepts no arguments
            if (rs.block.argAction == null) {
                antlrTool.warning("Rule '" + rr.targetRule +
                    "' accepts no arguments", grammar.getFilename(),
                    rr.getLine(), rr.getColumn());
            }
        }
        else {
            // For C++, no warning if rule has parameters, because there may be default
            // values for all of the parameters
            if (rs.block.argAction != null) {
                antlrTool.warning("Missing parameters on reference to rule " +
                    rr.targetRule, grammar.getFilename(),
                    rr.getLine(), rr.getColumn());
            }
        }
        // // _println(");");
        _println(")");

        // move down to the first child while parsing
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("_t = retTree_;");
            println("_t = retTree_");
        }
    }

    protected void genSemPred(String pred, int line) {
        // translate $ and # references
        ActionTransInfo tInfo = new ActionTransInfo();
        pred = processActionForSpecialSymbols(pred, line, currentRule, tInfo);
        // ignore translation info...we don't need to do anything with it.
        String escapedPred = charFormatter.escapeString(pred);

        // if debugging, wrap the semantic predicate evaluation in a method
        // that can tell SemanticPredicateListeners the result
        if (grammar.debuggingOutput &&
            ( (grammar instanceof ParserGrammar) ||
            (grammar instanceof LexerGrammar)))
            pred =
                "fireSemanticPredicateEvaluated(antlr.debug.SemanticPredicateEvent.VALIDATING,"
                + addSemPred(escapedPred) + "," + pred + ")";
        // // println("if (!(" + pred + "))");
        println("If (Not(" + pred + ")) Then ");
        tabadd();
        // // println("throw new SemanticException(\"" + escapedPred + "\");");
        println("Throw new SemanticException(\"" + escapedPred + "\")");
        tabsub();
        println("End If");
    }

    /** Write an array of Strings which are the semantic predicate
     *  expressions.  The debugger will reference them by number only
     */
    protected void genSemPredMap() {
        Enumeration e = semPreds.elements();
        // // println("private string[] _semPredNames = {");
        println("Private _semPredNames As String = {");
        tabadd();
        while (e.hasMoreElements())
            println("\"" + e.nextElement() + "\",");
        tabsub();
        // // println("};");
        println("}");
    }

    protected void genSynPred(SynPredBlock blk, String lookaheadExpr) {
        if (DEBUG_CODE_GENERATOR)
            System.out.println("gen=>(" + blk + ")");

            // Dump synpred result variable
        // // println("bool synPredMatched" + blk.ID + " = false;");
        println("Dim synPredMatched" + blk.ID + " As Boolean = False");
        // Gen normal lookahead test
        // // println("if (" + lookaheadExpr + ")");
        println("If (" + lookaheadExpr + ") Then" );
        // // println("{");
        tabadd();

        // Save input state
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("AST __t" + blk.ID + " = _t;");
            println("AST __t" + blk.ID + " = _t");
        }
        else {
            // // println("int _m" + blk.ID + " = mark();");
            println("int _m" + blk.ID + " = mark()");
        }

        // Once inside the try, assume synpred works unless exception caught
        // // println("synPredMatched" + blk.ID + " = true;");
        println("synPredMatched" + blk.ID + " = true");
        // // println("inputState.guessing++;");
        println("inputState.guessing += 1");

        // if debugging, tell listeners that a synpred has started
        if (grammar.debuggingOutput && ( (grammar instanceof ParserGrammar) ||
            (grammar instanceof LexerGrammar))) {
            // // println("fireSyntacticPredicateStarted();");
            println("fireSyntacticPredicateStarted()");
        }

        syntacticPredLevel++;
        // // println("try {");
        println("Try ");
        tabadd();
        gen( (AlternativeBlock) blk); // gen code to test predicate
        tabsub();
        //println("System.out.println(\"pred "+blk+" succeeded\");");
        // // println("}");
        //kunle: lose a few warnings cheaply
        //  println("catch (" + exceptionThrown + " pe)");
        println("Catch   pe As " + exceptionThrown );
        // // println("{");
        tabadd();
        // // println("synPredMatched" + blk.ID + " = false;");
        println("synPredMatched" + blk.ID + " = False");
        //println("System.out.println(\"pred "+blk+" failed\");");
        tabsub();
        // // println("}");
        println("End Try");

        // Restore input state
        if (grammar instanceof TreeWalkerGrammar) {
            // // println("_t = __t" + blk.ID + ";");
            println("_t = __t" + blk.ID );
        }
        else {
            // // println("rewind(_m" + blk.ID + ");");
            println("rewind(_m" + blk.ID + ")");
        }

        // // println("inputState.guessing--;");
        println("inputState.guessing -= 1");

        // if debugging, tell listeners how the synpred turned out
        if (grammar.debuggingOutput && ( (grammar instanceof ParserGrammar) ||
            (grammar instanceof LexerGrammar))) {
            // // println("if (synPredMatched" + blk.ID + ")");
            println("If (synPredMatched" + blk.ID + ") Then");
            tabadd();
            // // println("  fireSyntacticPredicateSucceeded();");
            println("fireSyntacticPredicateSucceeded()");
            tabsub();
            // // println("else");
            println("Else");
            tabadd();
            // // println("fireSyntacticPredicateFailed();");
            println("fireSyntacticPredicateFailed()");
            tabsub();
            println("End If");
        }

        syntacticPredLevel--;
        tabsub();

        // Close lookahead test
        // // println("}");
        println("End If"); // // "If (" + lookaheadExpr + ") Then"

        // Test synred result
        // // println("if ( synPredMatched" + blk.ID + " )");
        println("If ( synPredMatched" + blk.ID + " ) Then ");
        // // println("{");
    }

    /** Generate a static array containing the names of the tokens,
     * indexed by the token type values.  This static array is used
     * to format error messages so that the token identifers or literal
     * strings are displayed instead of the token numbers.
     *
     * If a lexical rule has a paraphrase, use it rather than the
     * token label.
     */
    public void genTokenStrings() {
        // Generate a string for each token.  This creates a static
        // array of Strings indexed by token type.
        println("");
        // // println("public static readonly string[] tokenNames_ = new string[] {");
        // F?
        println("Public Shared Readonly tokenNames_() As String = tokenNames_Extract()");
        // // println("Public Shared Readonly tokenNames_() As String = {");


        println("Public Shared Function tokenNames_Extract() As String()");
        tabadd();
        println("Dim sb As New StringBuilder(256)");
        println("Dim s_1 As String");
        // Walk the token vocabulary and generate a Vector of strings
        // from the tokens.
        Vector v = grammar.tokenManager.getVocabulary();
        for (int i = 0; i < v.size(); i++) {
            String s = (String) v.elementAt(i);
            if (s == null) {
                s = "<" + String.valueOf(i) + ">";
            }
            if (!s.startsWith("\"") && !s.startsWith("<")) {
                TokenSymbol ts = (TokenSymbol) grammar.tokenManager.
                    getTokenSymbol(s);
                if (ts != null && ts.getParaphrase() != null) {
                    s = StringUtils.stripFrontBack(ts.getParaphrase(), "\"",
                        "\"");
                }
            }
            else if (s.startsWith("\"")) {
                s = StringUtils.stripFrontBack(s, "\"", "\"");
            }
            // // print(charFormatter.literalString(s));
            println("s_1 = " + charFormatter.literalString(s));
            if ( i > 0 ) {println("sb.Append(\",\")");}
            println("sb.Append(s_1)");
            // // if (i != v.size() - 1) { _print(",");}
            // // _println("");
        }
        println("Return sb.toString().split(\",\")");
        // Close the string array initailizer
        tabsub();
        // // println("};");
        println("End Function");
    }





    /** Generate the token types vb file */
    protected void genTokenTypes(TokenManager tm) throws IOException {
        // Open the token output vb file and set the currentOutput stream
        // SAS: file open was moved to a method so a subclass can override
        //      This was mainly for the VAJ interface
        setupOutput(tm.getName() + TokenTypesFileSuffix);

        // // tabs = 0;
        tabclear();
        // Generate the header common to all vb files
        genHeader();
        // Do not use printAction because we assume tabs==0
        println(behavior.getHeaderAction(""));

        // Generate the vb namespace declaration (if specified)
        if (nameSpace != null)
            nameSpace.emitDeclarations(currentOutput);
        tabadd();
        println("");
        // Encapsulate the definitions in a class.  This has to be done as a class because
        // they are all constants and vb inteface  types cannot contain constants.
        // // println("public class " + tm.getName() + TokenTypesFileSuffix);
        println("Public Class " + tm.getName() + TokenTypesFileSuffix);
        //println("public class " + getTokenTypesClassName());
        // // println("{");
        tabadd();

        genTokenDefinitions(tm);

        // Close the interface
        tabsub();
        // // println("}");
        println("End Class");

        tabsub();
        // Generate the vb namespace closures (if required)
        if (nameSpace != null)
            nameSpace.emitClosures(currentOutput);

            // Close the tokens output file
        currentOutput.close();
        currentOutput = null;
        exitIfError();
    }

    protected void genTokenDefinitions(TokenManager tm) throws IOException {
        // Generate a definition for each token type
        Vector v = tm.getVocabulary();

        // Do special tokens manually
        // // println("public const int EOF = " + Token.EOF_TYPE + ";");
        println("Public Const EOF As Integer = " + Token.EOF_TYPE );
        // // println("public const int NULL_TREE_LOOKAHEAD = " + Token.NULL_TREE_LOOKAHEAD + ";");
        println("Public Const NULL_TREE_LOOKAHEAD As Integer = " + Token.NULL_TREE_LOOKAHEAD );

        for (int i = Token.MIN_USER_TYPE; i < v.size(); i++) {
            String s = (String) v.elementAt(i);
            if (s != null) {
                if (s.startsWith("\"")) {
                    // a string literal
                    StringLiteralSymbol sl = (StringLiteralSymbol) tm.
                        getTokenSymbol(s);
                    if (sl == null) {
                        antlrTool.panic("String literal " + s + " not in symbol table");
                    }
                    else if (sl.label != null) {
                        // // println("public const int " + sl.label + " = " + i + ";");
                        println("Public Const " + sl.label + " As Integer = " + i );
                    }
                    else {
                        String mangledName = mangleLiteral(s);
                        if (mangledName != null) {
                            // We were able to create a meaningful mangled token name
                            // // println("public const int " + mangledName + " = " + i + ";");
                            println("Public Const " + mangledName + " As Integer = " + i);
                            // if no label specified, make the label equal to the mangled name
                            sl.label = mangledName;
                        }
                        else {
                            // // println("// " + s + " = " + i);
                            println("' ' // " + s + " = " + i);
                        }
                    }
                }
                else if (!s.startsWith("<")) {
                    // // println("public const int " + s + " = " + i + ";");
                    println("Public Const " + s + " As Integer = " + i);
                }
            }
        }
        println("");
    }

    /** Process a string for an simple expression for use in xx/action.g
     * it is used to cast simple tokens/references to the right type for
     * the generated language. Basically called for every element in
     * the vector to getASTCreateString(vector V)
     * @param str A String.
     */
    public String processStringForASTConstructor(String str) {
        /*
             System.out.println("processStringForASTConstructor: str = "+str+
             ", custom = "+(new Boolean(usingCustomAST)).toString()+
             ", tree = "+(new Boolean((grammar instanceof TreeWalkerGrammar))).toString()+
             ", parser = "+(new Boolean((grammar instanceof ParserGrammar))).toString()+
                           ", notDefined = "+(new Boolean((!(grammar.tokenManager.tokenDefined(str))))).toString()
                           );
         */
        if (usingCustomAST &&
            ( (grammar instanceof TreeWalkerGrammar) ||
            (grammar instanceof ParserGrammar)) &&
            ! (grammar.tokenManager.tokenDefined(str))) {
            //System.out.println("processStringForASTConstructor: "+str+" with cast");
            // // return "(AST)" + str;
            return "DirectCast(" + str + ", AST)";
        }
        else {
            //System.out.println("processStringForASTConstructor: "+str);
            return str;
        }
    }

    /** Get a string for an expression to generate creation of an AST subtree.
     * @param v A Vector of String, where each element is an expression
     *          in the target language yielding an AST node.
     */
    public String getASTCreateString(Vector v) {
        if (v.size() == 0) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        // // buf.append("(" + labeledElementASTType +")astFactory.make( (new ASTArray(" + v.size() +"))");
        buf.append("DirectCast(astFactory.make( new ASTArray(" + v.size() +")), " + labeledElementASTType +")");
        for (int i = 0; i < v.size(); i++) {
            buf.append(".add(" + v.elementAt(i) + ")");
        }
        buf.append(")");
        return buf.toString();
    }

    /** Get a string for an expression to generate creating of an AST node
     * @param atom The grammar node for which you are creating the node
     * @param str The arguments to the AST constructor
     */
    public String getASTCreateString(GrammarAtom atom, String astCtorArgs) {
        String astCreateString = "astFactory.create(" + astCtorArgs + ")";

        if (atom == null)
            return getASTCreateString(astCtorArgs);
        else {
            if (atom.getASTNodeType() != null) {
                // this Atom was instantiated from a Token that had an "AST" option - associating
                // it with a specific heterogeneous AST type - applied to either:
                // 1) it's underlying TokenSymbol (in the "tokens {} section" or,
                // 2) a particular token reference in the grammar
                //
                // For option (1), we simply generate a cast to hetero-AST type
                // For option (2), we generate a call to factory.create(Token, ASTNodeType) and cast it too
                TokenSymbol ts = grammar.tokenManager.getTokenSymbol(atom.
                    getText());
                if ( (ts == null) ||
                    (ts.getASTNodeType() != atom.getASTNodeType())) {
                    // // astCreateString = "(" + atom.getASTNodeType() + ") astFactory.create(" + astCtorArgs + ", \"" + atom.getASTNodeType() + "\")";
                    astCreateString = "DirectCast(astFactory.create(" + astCtorArgs + ", \"" + atom.getASTNodeType() + "\"), " + atom.getASTNodeType() + ")";
                }
                else if ( (ts != null) && (ts.getASTNodeType() != null))
                    astCreateString = "(" + ts.getASTNodeType() + ") " +
                        astCreateString;
            }
            else if (usingCustomAST){
                // // astCreateString = "(" + labeledElementASTType + ") " + astCreateString;
                astCreateString = "DirectCast(" +astCreateString + ", "+ labeledElementASTType + ") " ;
            }
        }
        return astCreateString;
    }

    /** Returns a string expression that creates an AST node using the specified
     *  AST constructor argument string.
     *  Parses the first (possibly only) argument in the supplied AST ctor argument
     *	string to obtain the token type -- ctorID.
     *
     *  IF the token type is a valid token symbol AND
     *	   it has an associated AST node type     AND
     *	   this is not a #[ID, "T", "ASTType"] constructor
     *	THEN
     *	   generate a call to factory.create(ID, Text, token.ASTNodeType())
     *
     *  #[ID, "T", "ASTType"] constructors are mapped to astFactory.create(ID, "T", "ASTType")
     *
     *  The supported AST constructor forms are:
     *		#[ID]
     *		#[ID, "text"]
     *  	#[ID, "text", ASTclassname]	-- introduced in 2.7.2
     *
     * @param astCtorArgs The arguments to the AST constructor
     */
    public String getASTCreateString(String astCtorArgs) {
        // kunle: 19-Aug-2002
        // This AST creation string is almost certainly[*1] a manual tree construction request.
        // From the manual [I couldn't read ALL of the code ;-)], this can only be one of:
        // 1) #[ID]                     -- 'astCtorArgs' contains: 'ID'                     (without quotes)    or,
        // 2) #[ID, "T"]                -- 'astCtorArgs' contains: 'ID, "Text"'             (without single quotes) or,
        // kunle: 08-Dec-2002 - 2.7.2a6
        // 3) #[ID, "T", "ASTTypeName"] -- 'astCtorArgs' contains: 'ID, "T", "ASTTypeName"' (without single quotes)
        //
        // [*1]  In my tests, 'atom' was '== null' only for manual tree construction requests

        if (astCtorArgs == null) {
            astCtorArgs = "";
        }
        String astCreateString = "astFactory.create(" + astCtorArgs + ")";
        String ctorID = astCtorArgs;
        String ctorText = null;
        int commaIndex;
        boolean ctorIncludesCustomType = false; // Is this a #[ID, "t", "ASTType"] constructor?

        commaIndex = astCtorArgs.indexOf(',');
        if (commaIndex != -1) {
            ctorID = astCtorArgs.substring(0, commaIndex); // the 'ID'   portion of #[ID, "Text"]
            ctorText = astCtorArgs.substring(commaIndex + 1, astCtorArgs.length()); // the 'Text' portion of #[ID, "Text"]
            commaIndex = ctorText.indexOf(',');
            if (commaIndex != -1) {
                // This is an AST creation of the form: #[ID, "Text", "ASTTypename"]
                // Support for this was introduced with 2.7.2a6
                // create default type or (since 2.7.2) 3rd arg is classname
                ctorIncludesCustomType = true;
            }
        }
        TokenSymbol ts = grammar.tokenManager.getTokenSymbol(ctorID);
        if ( (null != ts) && (null != ts.getASTNodeType())) {
            // // astCreateString = "(" + ts.getASTNodeType() + ") " + astCreateString;
            astCreateString = "DirectCast(" + astCreateString + ", " + ts.getASTNodeType() + ")"  ;
        }
        else if (usingCustomAST) {
            // // astCreateString = "(" + labeledElementASTType + ") " + astCreateString;
            astCreateString = "DirectCast(" + astCreateString + ", " + labeledElementASTType + ")" ;
        }

        return astCreateString;
    }

    protected String getLookaheadTestExpression(Lookahead[] look, int k) {
        StringBuffer e = new StringBuffer(100);
        boolean first = true;

        e.append("(");
        for (int i = 1; i <= k; i++) {
            BitSet p = look[i].fset;
            if (!first) {
               // This is for the If conditions, must be changed to VB format, but for this function it is OK.
                e.append(") AndAlso (");
            }
            first = false;

            // Syn preds can yield <end-of-syn-pred> (epsilon) lookahead.
            // There is no way to predict what that token would be.  Just
            // allow anything instead.
            if (look[i].containsEpsilon()) {
                e.append("True");
            }
            else {
                e.append(getLookaheadTestTerm(i, p));
            }
        }
        e.append(")");

        return e.toString();
    }

    /**Generate a lookahead test expression for an alternate.  This
     * will be a series of tests joined by '&&' and enclosed by '()',
     * the number of such tests being determined by the depth of the lookahead.
     */
    protected String getLookaheadTestExpression(Alternative alt, int maxDepth) {
        int depth = alt.lookaheadDepth;
        if (depth == GrammarAnalyzer.NONDETERMINISTIC) {
            // if the decision is nondeterministic, do the best we can: LL(k)
            // any predicates that are around will be generated later.
            depth = grammar.maxk;
        }

        if (maxDepth == 0) {
            // empty lookahead can result from alt with sem pred
            // that can see end of token.  E.g., A : {pred}? ('a')? ;
            return "( True )";
        }
        // // return "(" + getLookaheadTestExpression(alt.cache, depth) + ")";
        // // This is for the If conditions, must be changed to VB format, but for this function it is OK.
        return "(" + getLookaheadTestExpression(alt.cache, depth) + ")";
    }

    /**Generate a depth==1 lookahead test expression given the BitSet.
     * This may be one of:
     * 1) a series of 'x==X||' tests
     * 2) a range test using >= && <= where possible,
     * 3) a bitset membership test for complex comparisons
     * @param k The lookahead level
     * @param p The lookahead set for level k
     */
    protected String getLookaheadTestTerm(int k, BitSet p) {
        // Determine the name of the item to be compared
        String ts = lookaheadString(k);

        // Generate a range expression if possible
        int[] elems = p.toArray();
        if (elementsAreRange(elems)) {
            return getRangeExpression(k, elems);
        }

        // Generate a bitset membership test if possible
        StringBuffer e;
        int degree = p.degree();
        if (degree == 0) {
            return "True";
        }

        if (degree >= bitsetTestThreshold) {
            int bitsetIdx = markBitsetForGen(p);
            return getBitsetName(bitsetIdx) + ".member(" + ts + ")";
        }

        // Otherwise, generate the long-winded series of "x==X||" tests
        e = new StringBuffer();
        for (int i = 0; i < elems.length; i++) {
            // Get the compared-to item (token or character value)
            String cs = getValueString(elems[i]);

            // Generate the element comparison
            if (i > 0) {
                // // e.append("||");
                e.append(" OrElse ");
            }
            e.append(ts);
            // // e.append("==");
            e.append(" = ");
            e.append(cs);
        }
        return e.toString();
    }

    /** Return an expression for testing a contiguous renage of elements
     * @param k The lookahead level
     * @param elems The elements representing the set, usually from BitSet.toArray().
     * @return String containing test expression.
     */
    public String getRangeExpression(int k, int[] elems) {
        if (!elementsAreRange(elems)) {
            antlrTool.panic("getRangeExpression called with non-range");
        }
        int begin = elems[0];
        int end = elems[elems.length - 1];

        // // return "(" + lookaheadString(k) + " >= " + getValueString(begin) + " && " + lookaheadString(k) + " <= " + getValueString(end) + ")";
        // // return "(" + lookaheadString(k) + " >= " + getValueString(begin) + " AndAlso " + lookaheadString(k) + " <= " + getValueString(end) + ")";
        return "(" + lookaheadString(k) + " >= " + this.convertCharLiteralPresentationToVB(getValueString(begin)) + " AndAlso " + lookaheadString(k) + " <= " + convertCharLiteralPresentationToVB(getValueString(end)) + ")";
    }

    /** getValueString: get a string representation of a token or char value
     * @param value The token or char value
     */
    private String getValueString(int value) {
        String cs;
        if (grammar instanceof LexerGrammar) {
            cs = charFormatter.literalChar(value);
        }
        else {
            TokenSymbol ts = grammar.tokenManager.getTokenSymbolAt(value);
            if (ts == null) {
                return "" + value; // return token type as string
                // antlrTool.panic("vocabulary for token type " + value + " is null");
            }
            String tId = ts.getId();
            if (ts instanceof StringLiteralSymbol) {
                // if string literal, use predefined label if any
                // if no predefined, try to mangle into LITERAL_xxx.
                // if can't mangle, use int value as last resort
                StringLiteralSymbol sl = (StringLiteralSymbol) ts;
                String label = sl.getLabel();
                if (label != null) {
                    cs = label;
                }
                else {
                    cs = mangleLiteral(tId);
                    if (cs == null) {
                        cs = String.valueOf(value);
                    }
                }
            }
            else {
                cs = tId;
            }
        }
        return cs;
    }

    /**Is the lookahead for this alt empty? */
    protected boolean lookaheadIsEmpty(Alternative alt, int maxDepth) {
        int depth = alt.lookaheadDepth;
        if (depth == GrammarAnalyzer.NONDETERMINISTIC) {
            depth = grammar.maxk;
        }
        for (int i = 1; i <= depth && i <= maxDepth; i++) {
            BitSet p = alt.cache[i].fset;
            if (p.degree() != 0) {
                return false;
            }
        }
        return true;
    }

    private String lookaheadString(int k) {
        if (grammar instanceof TreeWalkerGrammar) {
            return "_t.Type";
        }
        return "LA(" + k + ")";
    }

    /** Mangle a string literal into a meaningful token name.  This is
     * only possible for literals that are all characters.  The resulting
     * mangled literal name is literalsPrefix with the text of the literal
     * appended.
     * @return A string representing the mangled literal, or null if not possible.
     */
    private String mangleLiteral(String s) {
        String mangled = antlrTool.literalsPrefix;
        for (int i = 1; i < s.length() - 1; i++) {
            if (!Character.isLetter(s.charAt(i)) &&
                s.charAt(i) != '_') {
                return null;
            }
            mangled += s.charAt(i);
        }
        if (antlrTool.upperCaseMangledLiterals) {
            mangled = mangled.toUpperCase();
        }
        return mangled;
    }

    /** Map an identifier to it's corresponding tree-node variable.
     * This is context-sensitive, depending on the rule and alternative
     * being generated
     * @param idParam The identifier name to map
     * @return The mapped id (which may be the same as the input), or null if the mapping is invalid due to duplicates
     */
    public String mapTreeId(String idParam, ActionTransInfo transInfo) {
        // if not in an action of a rule, nothing to map.
        if (currentRule == null)
            return idParam;

        boolean in_var = false;
        String id = idParam;
        if (grammar instanceof TreeWalkerGrammar) {
            if (!grammar.buildAST) {
                in_var = true;
            }
            // If the id ends with "_in", then map it to the input variable
            else if (id.length() > 3 &&
                id.lastIndexOf("_in") == id.length() - 3) {
                // Strip off the "_in"
                id = id.substring(0, id.length() - 3);
                in_var = true;
            }
        }

        // Check the rule labels.  If id is a label, then the output
        // variable is label_AST, and the input variable is plain label.
        for (int i = 0; i < currentRule.labeledElements.size(); i++) {
            AlternativeElement elt = (AlternativeElement) currentRule.
                labeledElements.elementAt(i);
            if (elt.getLabel().equals(id)) {
                return in_var ? id : id + "_AST";
            }
        }

        // Failing that, check the id-to-variable map for the alternative.
        // If the id is in the map, then output variable is the name in the
        // map, and input variable is name_in
        String s = (String) treeVariableMap.get(id);
        if (s != null) {
            if (s == NONUNIQUE) {
                // There is more than one element with this id
                antlrTool.error("Ambiguous reference to AST element " + id +
                    " in rule " + currentRule.getRuleName());
                return null;
            }
            else if (s.equals(currentRule.getRuleName())) {
                // a recursive call to the enclosing rule is
                // ambiguous with the rule itself.
//				if( in_var )
//					System.out.println("returning null (rulename)");
                antlrTool.error("Ambiguous reference to AST element " + id +
                    " in rule " + currentRule.getRuleName());
                return null;
            }
            else {
                return in_var ? s + "_in" : s;
            }
        }

        // Failing that, check the rule name itself.  Output variable
        // is rule_AST; input variable is rule_AST_in (treeparsers).
        if (id.equals(currentRule.getRuleName())) {
            String r = in_var ? id + "_AST_in" : id + "_AST";
            if (transInfo != null) {
                if (!in_var) {
                    transInfo.refRuleRoot = r;
                }
            }
            return r;
        }
        else {
            // id does not map to anything -- return itself.
            return id;
        }
    }

    /** Given an element and the name of an associated AST variable,
     * create a mapping between the element "name" and the variable name.
     */
    private void mapTreeVariable(AlternativeElement e, String name) {
        // For tree elements, defer to the root
        if (e instanceof TreeElement) {
            mapTreeVariable( ( (TreeElement) e).root, name);
            return;
        }

        // Determine the name of the element, if any, for mapping purposes
        String elName = null;

        // Don't map labeled items
        if (e.getLabel() == null) {
            if (e instanceof TokenRefElement) {
                // use the token id
                elName = ( (TokenRefElement) e).atomText;
            }
            else if (e instanceof RuleRefElement) {
                // use the rule name
                elName = ( (RuleRefElement) e).targetRule;
            }
        }
        // Add the element to the tree variable map if it has a name
        if (elName != null) {
            if (treeVariableMap.get(elName) != null) {
                // Name is already in the map -- mark it as duplicate
                treeVariableMap.remove(elName);
                treeVariableMap.put(elName, NONUNIQUE);
            }
            else {
                treeVariableMap.put(elName, name);
            }
        }
    }

    /** Lexically process tree-specifiers in the action.
     *  This will replace #id and #(...) with the appropriate
     *  function calls and/or variables.
     */
    protected String processActionForSpecialSymbols(String actionStr,
        int line,
        RuleBlock currentRule,
        ActionTransInfo tInfo) {
        if (actionStr == null || actionStr.length() == 0)
            return null;

        // The action trans info tells us (at the moment) whether an
        // assignment was done to the rule's tree root.
        if (grammar == null)
            return actionStr;

        // see if we have anything to do...
        if ( (grammar.buildAST && actionStr.indexOf('#') != -1) ||
            grammar instanceof TreeWalkerGrammar ||
            ( (grammar instanceof LexerGrammar ||
            grammar instanceof ParserGrammar)
            && actionStr.indexOf('$') != -1)) {
            // Create a lexer to read an action and return the translated version
            // // antlr.actions.vb.ActionLexer lexer = new antlr.actions.vb.ActionLexer(actionStr, currentRule, this, tInfo);
            antlr.actions.vb.ActionLexer lexer = new antlr.actions.vb.ActionLexer(actionStr, currentRule, this, tInfo);

            lexer.setLineOffset(line);
            lexer.setFilename(grammar.getFilename());
            lexer.setTool(antlrTool);

            try {
                lexer.mACTION(true);
                actionStr = lexer.getTokenObject().getText();
                // System.out.println("action translated: "+actionStr);
                // System.out.println("trans info is "+tInfo);
            }
            catch (RecognitionException ex) {
                lexer.reportError(ex);
                return actionStr;
            }
            catch (TokenStreamException tex) {
                antlrTool.panic("Error reading action:" + actionStr);
                return actionStr;
            }
            catch (CharStreamException io) {
                antlrTool.panic("Error reading action:" + actionStr);
                return actionStr;
            }
        }
        return actionStr;
    }

    private void setupGrammarParameters(Grammar g) {
        if (g instanceof ParserGrammar ||
            g instanceof LexerGrammar ||
            g instanceof TreeWalkerGrammar
            ) {
            /* RK: options also have to be added to Grammar.java and for options
             * on the file level entries have to be defined in
             * DefineGrammarSymbols.java and passed around via 'globals' in antlrTool.java
             */
            if (antlrTool.nameSpace != null)
                nameSpace = new VBNameSpace(antlrTool.nameSpace.getName());
                //genHashLines = antlrTool.genHashLines;

                /* let grammar level options override filelevel ones...
                 */
            if (g.hasOption("namespace")) {
                Token t = g.getOption("namespace");
                if (t != null) {
                    // // nameSpace = new CSharpNameSpace(t.getText());
                    nameSpace = new VBNameSpace(t.getText());
                }
            }
            /*
                                     if( g.hasOption("genHashLines") ) {
                    Token t = g.getOption("genHashLines");
                    if( t != null )  {
                 String val = StringUtils.stripFrontBack(t.getText(),"\"","\"");
                            genHashLines = val.equals("True");
                    }
                                     }
             */
        }

        if (g instanceof ParserGrammar) {
            labeledElementASTType = "AST";
            if (g.hasOption("ASTLabelType")) {
                Token tsuffix = g.getOption("ASTLabelType");
                if (tsuffix != null) {
                    String suffix = StringUtils.stripFrontBack(tsuffix.getText(),
                        "\"", "\"");
                    if (suffix != null) {
                        usingCustomAST = true;
                        labeledElementASTType = suffix;
                    }
                }
            }
            labeledElementType = "Token ";
            // // labeledElementInit = "null";
            labeledElementInit = "Nothing";
            commonExtraArgs = "";
            commonExtraParams = "";
            commonLocalVars = "";
            lt1Value = "LT(1)";
            exceptionThrown = "RecognitionException";
            // // throwNoViable ="throw new NoViableAltException(LT(1), getFilename());";
            throwNoViable ="throw new NoViableAltException(LT(1), getFilename())";
        }
        else if (g instanceof LexerGrammar) {
            labeledElementType = "char ";
            labeledElementInit = "'\\0'"; // // !?!
            commonExtraArgs = "";
            // // commonExtraParams = "bool _createToken";
            commonExtraParams = "Dim _createToken As Boolean";
            // // commonLocalVars = "int _ttype; Token _token=null; int _begin=text.Length;";
            commonLocalVars = "Dim _ttype As Integer,  _token As Token = Nothing, _begin As Integer = text.Length";
            lt1Value = "LA(1)";
            exceptionThrown = "RecognitionException";
            // // throwNoViable = "throw new NoViableAltForCharException((char)LA(1), getFilename(), getLine(), getColumn());";
            throwNoViable = "Throw new NoViableAltForCharException(CType(LA(1),char), getFilename(), getLine(), getColumn())";
        }
        else if (g instanceof TreeWalkerGrammar) {
            labeledElementASTType = "AST";
            labeledElementType = "AST";
            if (g.hasOption("ASTLabelType")) {
                Token tsuffix = g.getOption("ASTLabelType");
                if (tsuffix != null) {
                    String suffix = StringUtils.stripFrontBack(tsuffix.getText(),
                        "\"", "\"");
                    if (suffix != null) {
                        usingCustomAST = true;
                        labeledElementASTType = suffix;
                        labeledElementType = suffix;
                    }
                }
            }
            if (!g.hasOption("ASTLabelType")) {
                g.setOption("ASTLabelType",
                    new Token(ANTLRTokenTypes.STRING_LITERAL, "AST"));
            }
            // // labeledElementInit = "null";
            labeledElementInit = "Nothing";
            commonExtraArgs = "_t";
            commonExtraParams = "AST _t";
            commonLocalVars = "";
            if (usingCustomAST){
                // // lt1Value = "(_t==ASTNULL) ? null : (" + labeledElementASTType + ")_t";
                lt1Value = "IIf (_t = ASTNULL), Nothing , DirectCast(_t, " + labeledElementASTType + "))";
            }
            else
                lt1Value = "_t";
            exceptionThrown = "RecognitionException";
            // // throwNoViable = "throw new NoViableAltException(_t);";
            throwNoViable = "Throw new NoViableAltException(_t)";
        }
        else {
            antlrTool.panic("Unknown grammar type");
        }
    }

    /** This method exists so a subclass, namely VAJCodeGenerator,
     *  can open the file in its own evil way.  JavaCodeGenerator
     *  simply opens a text file...
     */
    public void setupOutput(String className) throws IOException {
        // // currentOutput = antlrTool.openOutputFile(className + ".cs");
        currentOutput = antlrTool.openOutputFile(className + ".vb");
    }

    /** Helper method from Eric Smith's version of VBCodeGenerator.*/
    private static String OctalToUnicode(String str) {
        // only do any conversion if the string looks like "'\003'"
        if ( (4 <= str.length()) &&
            ('\'' == str.charAt(0)) &&
            ('\\' == str.charAt(1)) &&
            ( ('0' <= str.charAt(2)) && ('7' >= str.charAt(2))) &&
            ('\'' == str.charAt(str.length() - 1))) {
            // convert octal representation to decimal, then to hex
            Integer x = Integer.valueOf(str.substring(2, str.length() - 1), 8);

            return "'\\x" + Integer.toHexString(x.intValue()) + "'";
        }
        else {
            return str;
        }
    }

    /** Helper method that returns the name of the interface/class/enum type for
             token type constants.
     */
    public String getTokenTypesClassName() {
        TokenManager tm = grammar.tokenManager;
        return new String(tm.getName() + TokenTypesFileSuffix);
    }

    private void declareSaveIndexVariableIfNeeded() {
        if (!bSaveIndexCreated) {
            // // println("int _saveIndex = 0;");
            println("Dim _saveIndex As Integer = 0");
            bSaveIndexCreated = true;
        }
    }
}